# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from nebula2.fbthrift.transport.TTransport import *

import os
import socket
import sys
import warnings

if sys.version_info[0] >= 3:
    from io import BytesIO as StringIO
    from urllib import parse
    from http import client
    # pyre-fixme[11]: Annotation `parse` is not defined as a type.
    urlparse = parse
    urllib = parse
    # pyre-fixme[11]: Annotation `client` is not defined as a type.
    httplib = client
else:
    from cStringIO import StringIO
    import urlparse
    import httplib  # @manual
    import urllib

class THttpClient(TTransportBase):

    """Http implementation of TTransport base."""

    def __init__(self, uri_or_host, port=None, path=None, ssl_context=None):
        """THttpClient supports two different types constructor parameters.

        THttpClient(host, port, path) - deprecated
        THttpClient(uri)

        Only the second supports https."""

        if port is not None:
            warnings.warn(
                "Please use the THttpClient('http://host:port/path') syntax",
                DeprecationWarning,
                stacklevel=2)
            self.host = uri_or_host
            self.http_host = self.host
            self.port = port
            assert path
            self.path = path
            self.scheme = 'http'
        else:
            parsed = urlparse.urlparse(uri_or_host)
            self.scheme = parsed.scheme
            assert self.scheme in ('http', 'https')
            if self.scheme == 'http':
                self.port = parsed.port or httplib.HTTP_PORT
            elif self.scheme == 'https':
                self.port = parsed.port or httplib.HTTPS_PORT
            self.host = parsed.hostname
            self.http_host = parsed.netloc
            self.path = parsed.path
            if parsed.query:
                self.path += '?%s' % parsed.query
        self.__wbuf = StringIO()
        self.__http = None
        self.__timeout = None
        self.__custom_headers = None
        self.ssl_context = ssl_context

    def open(self):
        if self.scheme == 'http':
            self.__http = httplib.HTTPConnection(self.host, self.port,
                                                 timeout=self.__timeout)
        else:
            self.__http = httplib.HTTPSConnection(self.host, self.port,
                                                  context=self.ssl_context,
                                                  timeout=self.__timeout)

    def close(self):
        self.__http.close()
        self.__http = None

    def isOpen(self):
        return self.__http is not None

    def setTimeout(self, ms):
        if ms is None:
            self.__timeout = None
        else:
            self.__timeout = ms / 1000.0

    def setCustomHeaders(self, headers):
        self.__custom_headers = headers

    def setCustomHeader(self, name, value):
        if self.__custom_headers is None:
            self.__custom_headers = {}
        self.__custom_headers[name] = value

    def read(self, sz):
        return self.response.read(sz)

    def write(self, buf):
        self.__wbuf.write(buf)

    def flush(self):
        if self.isOpen():
            self.close()
        self.open()

        # Pull data out of buffer
        data = self.__wbuf.getvalue()
        self.__wbuf = StringIO()

        # HTTP request
        self.__http.putrequest('POST', self.path, skip_host=True)

        if not self.__custom_headers or 'Host' not in self.__custom_headers:
            self.__http.putheader('Host', self.http_host)

        self.__http.putheader('Content-Type', 'application/x-thrift')
        self.__http.putheader('Content-Length', str(len(data)))

        if not self.__custom_headers or 'User-Agent' not in \
                self.__custom_headers:
            user_agent = 'Python/THttpClient'
            script = os.path.basename(sys.argv[0])
            if script:
                user_agent = '%s (%s)' % (user_agent, urllib.quote(script))
            self.__http.putheader('User-Agent', user_agent)

        if self.__custom_headers:
            if sys.version_info[0] >= 3:
                custom_headers_iter = self.__custom_headers.items()
            else:
                custom_headers_iter = self.__custom_headers.items()
            for key, val in custom_headers_iter:
                self.__http.putheader(key, val)

        try:
            self.__http.endheaders()

            # Write payload
            self.__http.send(data)
        except socket.gaierror as e:
            raise TTransportException(TTransportException.NOT_OPEN, str(e))
        except Exception as e:
            raise TTransportException(TTransportException.UNKNOWN, str(e))


        # Get reply to flush the request
        self.response = self.__http.getresponse()
        self.code = self.response.status
        self.headers = self.response.getheaders()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from nebula2.fbthrift.protocol.TProtocol import TProtocolBase, TProtocolException
from nebula2.fbthrift.Thrift import TApplicationException, TMessageType
from nebula2.fbthrift.protocol.TBinaryProtocol import TBinaryProtocolAccelerated
from nebula2.fbthrift.protocol.TCompactProtocol import TCompactProtocolAccelerated
from nebula2.fbthrift.transport.THeaderTransport import THeaderTransport, CLIENT_TYPE


class THeaderProtocolAccelerate(object):

    """Base class for pass through header protocols"""
    def get_protocol_id(self):
        raise NotImplementedError()


class THeaderProtocol(TProtocolBase, THeaderProtocolAccelerate):

    """Pass through header protocol (transport can set)"""
    T_BINARY_PROTOCOL = 0
    T_JSON_PROTOCOL = 1
    T_COMPACT_PROTOCOL = 2

    __proto = None
    __proto_id = None

    def get_protocol_id(self):
        return self.__proto_id

    def reset_protocol(self):
        if self.__proto_id == self.trans.get_protocol_id():
            return

        proto_id = self.trans.get_protocol_id()

        if proto_id == self.T_BINARY_PROTOCOL:
            self.__proto = TBinaryProtocolAccelerated(self.trans,
                    self.strictRead, True)
        elif proto_id == self.T_COMPACT_PROTOCOL:
            self.__proto = TCompactProtocolAccelerated(self.trans)
        else:
            raise TApplicationException(TProtocolException.INVALID_PROTOCOL,
                                        "Unknown protocol requested")
        self.__proto_id = proto_id

    def __init__(self, trans, strictRead=False,
                 client_types=None, client_type=None):
        """Create a THeaderProtocol instance

        @param transport(TTransport) The underlying transport.
        @param strictRead(bool) Turn on strictRead if using TBinaryProtocol
        @param client_types([CLIENT_TYPE.HEADER, ...])
                   List of client types to support.  Defaults to
                   CLIENT_TYPE.HEADER only.
        """

        if isinstance(trans, THeaderTransport):
            trans._THeaderTransport__supported_client_types = set(
                client_types or (CLIENT_TYPE.HEADER,))
            if client_type is not None:
                trans._THeaderTransport__client_type = client_type
            htrans = trans
        else:
            htrans = THeaderTransport(trans, client_types, client_type)
        TProtocolBase.__init__(self, htrans)
        self.strictRead = strictRead
        self.reset_protocol()

    def writeMessageBegin(self, name, type, seqid):
        self.__proto.writeMessageBegin(name, type, seqid)
        if type == TMessageType.CALL or type == TMessageType.ONEWAY:
            # All client to server coms should have a unique seq_id in HEADER
            self.trans.seq_id = seqid

    def writeMessageEnd(self):
        self.__proto.writeMessageEnd()

    def writeStructBegin(self, name):
        self.__proto.writeStructBegin(name)

    def writeStructEnd(self):
        self.__proto.writeStructEnd()

    def writeFieldBegin(self, name, type, id):
        self.__proto.writeFieldBegin(name, type, id)

    def writeFieldEnd(self):
        self.__proto.writeFieldEnd()

    def writeFieldStop(self):
        self.__proto.writeFieldStop()

    def writeMapBegin(self, ktype, vtype, size):
        self.__proto.writeMapBegin(ktype, vtype, size)

    def writeMapEnd(self):
        self.__proto.writeMapEnd()

    def writeListBegin(self, etype, size):
        self.__proto.writeListBegin(etype, size)

    def writeListEnd(self):
        self.__proto.writeListEnd()

    def writeSetBegin(self, etype, size):
        self.__proto.writeSetBegin(etype, size)

    def writeSetEnd(self):
        self.__proto.writeSetEnd()

    def writeBool(self, bool):
        self.__proto.writeBool(bool)

    def writeByte(self, byte):
        self.__proto.writeByte(byte)

    def writeI16(self, i16):
        self.__proto.writeI16(i16)

    def writeI32(self, i32):
        self.__proto.writeI32(i32)

    def writeI64(self, i64):
        self.__proto.writeI64(i64)

    def writeDouble(self, dub):
        self.__proto.writeDouble(dub)

    def writeFloat(self, flt):
        self.__proto.writeFloat(flt)

    def writeString(self, str):
        self.__proto.writeString(str)

    def readMessageBegin(self):
        #Read the next frame, and change protocols if needed
        try:
            self.trans._reset_protocol()
            self.reset_protocol()
        except TApplicationException as ex:
            if self.__proto:
                self.writeMessageBegin(b"", TMessageType.EXCEPTION, 0)
                ex.write(self)
                self.writeMessageEnd()
                self.trans.flush()
        return self.__proto.readMessageBegin()

    def readMessageEnd(self):
        return self.__proto.readMessageEnd()

    def readStructBegin(self):
        return self.__proto.readStructBegin()

    def readStructEnd(self):
        return self.__proto.readStructEnd()

    def readFieldBegin(self):
        return self.__proto.readFieldBegin()

    def readFieldEnd(self):
        return self.__proto.readFieldEnd()

    def readMapBegin(self):
        return self.__proto.readMapBegin()

    def readMapEnd(self):
        return self.__proto.readMapEnd()

    def readListBegin(self):
        return self.__proto.readListBegin()

    def readListEnd(self):
        return self.__proto.readListEnd()

    def readSetBegin(self):
        return self.__proto.readSetBegin()

    def readSetEnd(self):
        return self.__proto.readSetEnd()

    def readBool(self):
        return self.__proto.readBool()

    def readByte(self):
        return self.__proto.readByte()

    def readI16(self):
        return self.__proto.readI16()

    def readI32(self):
        return self.__proto.readI32()

    def readI64(self):
        return self.__proto.readI64()

    def readDouble(self):
        return self.__proto.readDouble()

    def readFloat(self):
        return self.__proto.readFloat()

    def readString(self):
        return self.__proto.readString()

class THeaderProtocolFactory(object):
    def __init__(self, strictRead=False, client_types=None, client_type=None):
        self.strictRead = strictRead
        self.client_types = client_types
        self.client_type = client_type

    def getProtocol(self, trans):
        prot = THeaderProtocol(trans, self.strictRead, self.client_types,
                self.client_type)
        return prot



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from nebula2.fbthrift.Thrift import TType

def fix_spec(all_structs):
    for s in all_structs:
        spec = s.thrift_spec
        for t in spec:
            if t is None:
                continue
            elif t[1] == TType.STRUCT:
                t[3][1] = t[3][0].thrift_spec
            elif t[1] in (TType.LIST, TType.SET):
                _fix_list_or_set(t[3])
            elif t[1] == TType.MAP:
                _fix_map(t[3])

def _fix_list_or_set(element_type):
    if element_type[0] == TType.STRUCT:
        element_type[1][1] = element_type[1][0].thrift_spec
    elif element_type[0] in (TType.LIST, TType.SET):
        _fix_list_or_set(element_type[1])
    elif element_type[0] == TType.MAP:
        _fix_map(element_type[1])

def _fix_map(element_type):
    if element_type[0] == TType.STRUCT:
        element_type[1][1] = element_type[1][0].thrift_spec
    elif element_type[0] in (TType.LIST, TType.SET):
        _fix_list_or_set(element_type[1])
    elif element_type[0] == TType.MAP:
        _fix_map(element_type[1])

    if element_type[2] == TType.STRUCT:
        element_type[3][1] = element_type[3][0].thrift_spec
    elif element_type[2] in (TType.LIST, TType.SET):
        _fix_list_or_set(element_type[3])
    elif element_type[2] == TType.MAP:
        _fix_map(element_type[3])



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys

if sys.version_info[0] >= 3:

    from io import BytesIO

    class BytesStrIO(BytesIO):
        def __init__(self, *args):
            args_new = []
            for arg in args:
                if not isinstance(arg, (bytes, memoryview)):
                    args_new.append(arg.encode())
                else:
                    args_new.append(arg)
            BytesIO.__init__(self, *args_new)

        def write(self, data):
            if isinstance(data, (bytes, memoryview)):
                BytesIO.write(self, data)
            else:
                BytesIO.write(self, data.encode())



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os.path
import socket
import tempfile
import threading
import time
import unittest

import nebula2.fbthrift.transport.TSocket as TSocket
import nebula2.fbthrift.transport.TTransport as TTransport


class TSocketTest(unittest.TestCase):

    def test_usage_as_context_manager(self):
        """
        Asserts that both TSocket and TServerSocket can be used with `with` and
        that their resources are disposed of at the close of the `with`.
        """
        text = b"hi"  # sample text to send over the wire
        with TSocket.TServerSocket(port=0, family=socket.AF_INET6) as server:
            addr = server.getSocketNames()[0]
            with TSocket.TSocket(host=addr[0], port=addr[1]) as conn:
                conn.write(text)
            self.assertFalse(conn.isOpen())
            with server.accept() as client:
                read = client.read(len(text))
            self.assertFalse(conn.isOpen())
        self.assertFalse(server.isListening())
        self.assertEquals(read, text)

    def test_server_context_errors(self):
        # Make sure the TServerSocket context manager doesn't
        # swallow exceptions
        def do_test():
            with TSocket.TServerSocket(port=0, family=socket.AF_INET6):
                raise Exception('test_error')

        self.assertRaisesRegexp(Exception, 'test_error', do_test)

    def test_open_failure(self):
        # Bind a server socket to an address, but don't actually listen on it.
        server_socket = socket.socket(socket.AF_INET6)
        try:
            server_socket.bind(('::', 0))
            server_port = server_socket.getsockname()[1]

            # Explicitly use "localhost" as the hostname, so that the
            # connect code will try both IPv6 and IPv4.  We want to
            # exercise the failure behavior when trying multiple addresses.
            sock = TSocket.TSocket(host='localhost', port=server_port)
            sock.setTimeout(50)  # ms
            try:
                sock.open()
                self.fail('unexpectedly succeeded to connect to closed socket')
            except TTransport.TTransportException:
                # sock.open() should not leave the file descriptor open
                # when it fails
                self.assertEquals(None, sock.handle)
                self.assertEquals({}, sock.handles)

                # Calling close() again on the socket should be a no-op,
                # and shouldn't throw an error
                sock.close()
        finally:
            server_socket.close()

    def test_poller_process(self):
        # Make sure that pollers do not fail when they're given None as timeout
        text = "hi"  # sample text to send over the wire
        with TSocket.TServerSocket(port=0, family=socket.AF_INET6) as server:
            addr = server.getSocketNames()[0]

            def write_data():
                # delay writing to verify that poller.process is waiting
                time.sleep(1)
                with TSocket.TSocket(host=addr[0], port=addr[1]) as conn:
                    conn.write(text)

            poller = TSocket.ConnectionSelect()
            thread = threading.Thread(target=write_data)
            thread.start()
            for filenos in server.handles.keys():
                poller.read(filenos)

            r, _, x = poller.process(timeout=None)

            thread.join()
            # Verify that r is non-empty
            self.assertTrue(r)

    def test_deprecated_str_form_of_port(self):
        # Make sure that the deprecated form of the `port` parameter is
        # accepted in TServerSocket and TSocket.
        port = "0"
        text = b"hi"  # sample text to send over the wire
        # NB: unfortunately unittest.TestCase.assertWarns isn't available until
        # py3.
        with TSocket.TServerSocket(port=port, family=socket.AF_INET6) as server:
            addr = server.getSocketNames()[0]
            with TSocket.TSocket(host=addr[0], port=str(addr[1])) as conn:
                conn.write(text)
            with server.accept() as client:
                read = client.read(len(text))
            self.assertEquals(read, text)

    def test_bad_port(self):
        port = 'bogus'
        with self.assertRaises(ValueError):
            with TSocket.TServerSocket(port=port):
                pass

        with self.assertRaises(ValueError):
            with TSocket.TSocket(port=port):
                pass

    def test_unix_socket(self):
        text = b"hi"  # sample text to send over the wire
        with tempfile.NamedTemporaryFile(delete=True) as fh:
            unix_socket = fh.name
            with TSocket.TServerSocket(unix_socket=unix_socket) as server:
                with TSocket.TSocket(unix_socket=unix_socket) as conn:
                    conn.write(text)
                with server.accept() as client:
                    read = client.read(len(text))
                self.assertEquals(read, text)
            # The socket will not be cleaned up when the server has been shutdown.
            self.assertTrue(os.path.exists(unix_socket))



#
# Autogenerated by Thrift
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#  @generated
#
__all__ = ['ttypes', 'constants', 'GraphStorageService', 'StorageAdminService', 'GeneralStorageService', 'InternalStorageService']



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def create_ThriftUnicodeDecodeError_from_UnicodeDecodeError(error, field_name):
    if isinstance(error, ThriftUnicodeDecodeError):
        error.field_names.append(field_name)
        return error
    return ThriftUnicodeDecodeError(
        error.encoding, error.object, error.start, error.end, error.reason, field_name
    )


class ThriftUnicodeDecodeError(UnicodeDecodeError):
    def __init__(self, encoding, object, start, end, reason, field_name):
        super(ThriftUnicodeDecodeError, self).__init__(encoding, object, start, end, reason)
        self.field_names = [field_name]

    def __str__(self):
        return "{error} when decoding field '{field}'".format(
            error=super(ThriftUnicodeDecodeError, self).__str__(), field="->".join(reversed(self.field_names))
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nebula2.fbthrift.Thrift import TType

import inspect
import sys
from typing import Any, Dict, List, Optional, Tuple, Type, cast
from types import ModuleType

ExtraTypeSpec = Any
FieldSpec = Tuple[
    int,  # field id
    TType,
    str,  # field_name
    ExtraTypeSpec,
    Any,  # default value
    int,  # 0=required, 1=optional, 2=default
]

_args_suffix = "_args"
_result_suffix = "_result"


class NoSuchFunctionError(ValueError):
    pass


class Function:
    """A class describing the argument types and return type of a thrift function."""

    def __init__(
        self,
        name: str,
        module: ModuleType,
        arg_specs: List[FieldSpec],
        result_spec: Optional[Tuple[FieldSpec]],
    ) -> None:
        self.name = name
        self.module = module
        self.arg_specs = arg_specs
        self.result_spec = result_spec

    @property
    def service_name(self) -> str:
        return self.module.__name__

    def __str__(self) -> str:
        arg_info = []
        for info in self.arg_specs:
            if info is None:
                # The thrift_spec format is pretty dumb, and attempts to make the tuple
                # indices match up with the field ID numbers in most cases (when the
                # field IDs are positive).  It fills gaps in the field IDs with None
                # elements.
                continue
            (
                _field_id,
                thrift_type,  # enum describing thrift base type
                arg_name,
                extra_type_spec,
                _default_value,
                _required,
            ) = info
            type_name = get_type_name(thrift_type, extra_type_spec)
            arg_info.append(f"{arg_name}: {type_name}")

        args_str = ", ".join(arg_info)
        if self.result_spec is None:
            # This is a oneway function.  The server sends no response at all (and
            # therefore cannot even indicate errors).  There is no confirmation that the
            # request was even received or processed.
            return f"{self.name}({args_str}) [oneway]"
        else:
            result_str = _get_result_str(self.result_spec)
            return f"{self.name}({args_str}) -> {result_str}"


def get_service_module_hierarchy(service_module: ModuleType) -> List[ModuleType]:
    """Given a thrift service module, return a list containing it plus all
    of the modules for its parent services.
    """
    iface_class = getattr(service_module, "Iface", None)
    if iface_class is None:
        raise TypeError(
            f"{service_module!r} does not look like a thrift service module"
        )

    result = []
    for cls in iface_class.__mro__:
        module = sys.modules[cls.__module__]
        # Check to make sure this actually looks like a thrift service module.
        # The MRO hierarchy will include the base "object" class which is not from a
        # thrift module.
        if hasattr(module, "Iface"):
            result.append(module)

    return result


def list_all_functions(service_module: ModuleType) -> Dict[str, Function]:
    """Given a thrift service module, return information about the functions exposed by
    this service, including all of its inherited functions.
    """
    modules = get_service_module_hierarchy(service_module)
    functions: Dict[str, Function] = {}
    for module in reversed(modules):
        functions.update(list_service_functions(module))

    return functions


def list_service_functions(service_module: ModuleType) -> Dict[str, Function]:
    """Given a thrift service module, return information about the functions it exposes.
    This returns only the functions directly defined by this service, excluding
    functions inherited from parent service interfaces.  Use list_all_functions() if you
    also want inherited functions to be returned.
    """
    functions = {}
    for name in dir(service_module):
        if not name.endswith(_args_suffix):
            continue
        fn_name = name[: -len(_args_suffix)]
        functions[fn_name] = get_function_info(service_module, fn_name)

    return functions


def get_function_info(
    service_module: ModuleType, fn_name: str, search_parent_interfaces: bool = True
) -> Function:
    """Return a Function object describing the interface of the specified function
    in the supplied thrift service module.
    """
    if search_parent_interfaces:
        modules = get_service_module_hierarchy(service_module)
    else:
        modules = [service_module]

    for module in modules:
        fn_info = _get_function_info(module, fn_name)
        if fn_info is not None:
            return fn_info

    raise NoSuchFunctionError(f"no function named {fn_name}")


def _get_function_info(service_module: ModuleType, fn_name: str) -> Optional[Function]:
    args_type = getattr(service_module, fn_name + _args_suffix, None)
    if args_type is None:
        # This function is not defined by this module
        return None

    ordered_arg_specs = _get_function_arg_specs(service_module, fn_name)
    result_type = getattr(service_module, fn_name + _result_suffix, None)

    result_spec = (
        None if result_type is None else cast(Tuple[FieldSpec], result_type.thrift_spec)
    )
    return Function(fn_name, service_module, ordered_arg_specs, result_spec)


def _get_function_arg_specs(
    service_module: ModuleType, fn_name: str
) -> List[FieldSpec]:
    args_type = getattr(service_module, fn_name + _args_suffix)

    # Re-order the argument types to actually match the python function argument
    # ordering.
    #
    # The entries in the thrift spec will be ordered by thrift field ID numbers (and
    # it actually tries to match up the tuple indices to the thrift field IDs, which
    # results in huge tuples if you happen to use a large ID value).
    #
    # The field ID numbers do not necessarily match the source code order of the
    # arguments.
    iface_class = getattr(service_module, "Iface")
    py_fn = getattr(iface_class, fn_name)
    sig = inspect.signature(py_fn)

    # First build a dictionary of the thrift specs by name
    arg_specs = {}
    for spec in args_type.thrift_spec:
        if spec is None:
            continue
        arg_name = spec[2]
        arg_specs[arg_name] = spec

    # Now put them in the correct order
    ordered_arg_specs = []
    for param in list(sig.parameters)[1:]:  # skip the initial "self" parameter
        arg_spec = arg_specs.pop(param)
        ordered_arg_specs.append(arg_spec)

    assert (
        not arg_specs
    ), f"did not consume all thrift arguments for {fn_name}: {arg_specs}"

    return ordered_arg_specs


def get_type_name(thrift_type: TType, extra_type_spec: Any) -> str:
    """Convert information about a thrift type into a human-readable string."""
    if thrift_type == TType.STRING:
        is_binary = not cast(bool, extra_type_spec)
        if is_binary:
            return "binary"
        else:
            return "string"
    elif thrift_type == TType.BOOL:
        return "bool"
    elif thrift_type == TType.BYTE:
        return "byte"
    elif thrift_type == TType.DOUBLE:
        return "double"
    elif thrift_type == TType.I16:
        return "i16"
    elif thrift_type == TType.I32:
        if extra_type_spec is None:
            return "i32"
        else:
            # This is an enum, and the extra_type_spec contains the enum type
            python_type = cast(Type, extra_type_spec)
            return f"{python_type.__module__}.{python_type.__qualname__}"
    elif thrift_type == TType.I64:
        return "i64"
    elif thrift_type == TType.DOUBLE:
        return "double"
    elif thrift_type == TType.FLOAT:
        return "float"
    elif thrift_type == TType.STRUCT:
        python_type, type_spec, is_union = extra_type_spec
        return f"{python_type.__module__}.{python_type.__qualname__}"
    elif thrift_type == TType.MAP:
        (
            key_thrift_type,
            key_extra_spec,
            value_thrift_type,
            value_extra_spec,
        ) = extra_type_spec
        key_type_name = get_type_name(key_thrift_type, key_extra_spec)
        value_type_name = get_type_name(value_thrift_type, value_extra_spec)
        return f"map<{key_type_name}, {value_type_name}>"
    elif thrift_type == TType.SET:
        (elem_thrift_type, elem_extra_spec) = extra_type_spec
        elem_type_name = get_type_name(elem_thrift_type, elem_extra_spec)
        return f"set<{elem_type_name}>"
    elif thrift_type == TType.LIST:
        (elem_thrift_type, elem_extra_spec) = extra_type_spec
        elem_type_name = get_type_name(elem_thrift_type, elem_extra_spec)
        return f"list<{elem_type_name}>"
    else:
        # The thrift compiler currently does not emit any other
        # types for python.  (C++ also has a "STREAM" type but this is not
        # currently supported for python.)
        return "unsupported thrift type"


def _get_result_str(result_spec: Tuple[FieldSpec]) -> str:
    # The entry with field ID 0 is the success result.  If the function returns void
    # this will not be present.
    # All other entries are possible exception types that may be thrown.
    for info in result_spec:
        if info is None:
            continue
        (
            field_id,
            thrift_type,  # enum describing thrift base type
            _arg_name,
            extra_type_spec,
            _default_value,
            _required,
        ) = info
        if field_id == 0:
            return get_type_name(thrift_type, extra_type_spec)
    return "void"



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from nebula2.fbthrift.Thrift import *

class TProtocolException(TException):

    """Custom Protocol Exception class"""

    UNKNOWN = 0
    INVALID_DATA = 1
    NEGATIVE_SIZE = 2
    SIZE_LIMIT = 3
    BAD_VERSION = 4
    INVALID_PROTOCOL = 5
    MISSING_REQUIRED_FIELD = 6

    def __init__(self, type=UNKNOWN, message=None):
        TException.__init__(self, message)
        self.type = type

class TProtocolBase:

    """Base class for Thrift protocol driver."""

    def __init__(self, trans):
        self.trans = trans

    def writeMessageBegin(self, name, ttype, seqid):
        pass

    def writeMessageEnd(self):
        pass

    def writeStructBegin(self, name):
        pass

    def writeStructEnd(self):
        pass

    def writeUnionBegin(self, name):
        self.writeStructBegin(name)

    def writeUnionEnd(self):
        self.writeStructEnd()

    def writeFieldBegin(self, name, type, id):
        pass

    def writeFieldEnd(self):
        pass

    def writeFieldStop(self):
        pass

    def writeMapBegin(self, ktype, vtype, size):
        pass

    def writeMapEnd(self):
        pass

    def writeListBegin(self, etype, size):
        pass

    def writeListEnd(self):
        pass

    def writeSetBegin(self, etype, size):
        pass

    def writeSetEnd(self):
        pass

    def writeBool(self, bool_val):
        pass

    def writeByte(self, byte):
        pass

    def writeI16(self, i16):
        pass

    def writeI32(self, i32):
        pass

    def writeI64(self, i64):
        pass

    def writeDouble(self, dub):
        pass

    def writeFloat(self, flt):
        pass

    def writeString(self, str):
        pass

    def readMessageBegin(self):
        pass

    def readMessageEnd(self):
        pass

    def readStructBegin(self):
        pass

    def readStructEnd(self):
        pass

    def readFieldBegin(self):
        pass

    def readFieldEnd(self):
        pass

    def readMapBegin(self):
        pass

    def readMapEnd(self):
        pass

    def readListBegin(self):
        pass

    def readListEnd(self):
        pass

    def readSetBegin(self):
        pass

    def readSetEnd(self):
        pass

    def readBool(self):
        pass

    def readByte(self):
        pass

    def readI16(self):
        pass

    def readI32(self):
        pass

    def readI64(self):
        pass

    def readDouble(self):
        pass

    def readFloat(self):
        pass

    def readString(self):
        pass

    def skip(self, type):
        if type == TType.BOOL:
            self.readBool()
        elif type == TType.BYTE:
            self.readByte()
        elif type == TType.I16:
            self.readI16()
        elif type == TType.I32:
            self.readI32()
        elif type == TType.I64:
            self.readI64()
        elif type == TType.DOUBLE:
            self.readDouble()
        elif type == TType.FLOAT:
            self.readFloat()
        elif type == TType.STRING:
            self.readString()
        elif type == TType.STRUCT:
            name = self.readStructBegin()
            while True:
                (name, type, id) = self.readFieldBegin()
                if type == TType.STOP:
                    break
                self.skip(type)
                self.readFieldEnd()
            self.readStructEnd()
        elif type == TType.MAP:
            (ktype, vtype, size) = self.readMapBegin()
            for _ in range(size):
                self.skip(ktype)
                self.skip(vtype)
            self.readMapEnd()
        elif type == TType.SET:
            (etype, size) = self.readSetBegin()
            for _ in range(size):
                self.skip(etype)
            self.readSetEnd()
        elif type == TType.LIST:
            (etype, size) = self.readListBegin()
            for _ in range(size):
                self.skip(etype)
            self.readListEnd()
        else:
            raise TProtocolException(
                TProtocolException.INVALID_DATA,
                "Unexpected type for skipping {}".format(type)
            )

    def readIntegral(self, type):
        if type == TType.BOOL:
            return self.readBool()
        elif type == TType.BYTE:
            return self.readByte()
        elif type == TType.I16:
            return self.readI16()
        elif type == TType.I32:
            return self.readI32()
        elif type == TType.I64:
            return self.readI64()
        else:
            raise Exception("Unknown integral type: %s" % str(type))

    def readFloatingPoint(self, type):
        if type == TType.FLOAT:
            return self.readFloat()
        elif type == TType.DOUBLE:
            return self.readDouble()
        else:
            raise Exception("Unknown floating point type: %s" % str(type))

class TProtocolFactory:
    def getProtocol(self, trans):
        pass



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

"""
PyRemote

Used by PyRemote files generated by
  /thrift/compiler/generate/t_py_generator.cc

Remote.run is the interface used by the generated code.
Based on whether --host or --url is specified as a commandline option,
either a RemoteHostClient or RemoteHttpClient is instantiated to
handle the request.

Additional remote client types (subclasses of RemoteClient) can be
registered with the Remote class to define different ways of specifying a
host or communicating with the host. When registering a new client type,
you can specify the option used to select that type (i.e., url) with the
SELECTOR_OPTIONS attribute, and you can specify additional commandline options
with the CMDLINE_OPTIONS attribute. See the implementations of RemoteHostClient
and RemoteHttpClient for examples.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import json
import os
import pprint
import sys
import traceback

from six.moves.urllib.parse import urlparse
from six import string_types
from nebula2.fbthrift import Thrift
from nebula2.fbthrift.transport import TTransport, TSocket, TSSLSocket, THttpClient
from nebula2.fbthrift.transport.THeaderTransport import THeaderTransport
from nebula2.fbthrift.protocol import TBinaryProtocol, TCompactProtocol, \
    TJSONProtocol, THeaderProtocol, TSimpleJSONProtocol


class Function(object):
    """Metadata for a service method"""
    def __init__(self, fn_name, svc_name, return_type, args):
        self.fn_name = fn_name
        self.svc_name = svc_name
        self.return_type = return_type
        self.args = args


def print_functions(functions, service_names, out, local_only=False):
    """Print all the functions available from this thrift service"""
    fns_by_service_name = {svc_name: {} for svc_name in service_names}
    for fn in functions.values():
        fns_by_service_name[fn.svc_name][fn.fn_name] = fn

    svc_names = service_names[0:1] if local_only else \
                    reversed(service_names)
    for svc_name in svc_names:
        out.write('Functions in %s:\n' % (svc_name,))
        for fn_name, fn in sorted(fns_by_service_name[svc_name].items()):
            if fn.return_type is None:
                out.write('  oneway void ')
            else:
                out.write('  %s ' % (fn.return_type,))
            out.write(fn_name + '(')
            out.write(', '.join('%s %s' % (type, name)
                                for type, name, true_type in fn.args))
            out.write(')\n')


format_to_helper = {
    "input": {},
    "output": {},
}
format_to_help_message = {
    "input": {},
    "output": {},
}


def add_format(name, format_type, help_msg=None):
    """
    Decorate function to set it as a handler for the specified format and format_type

    All functions with same format_type must share the same interface/signature.
    In other cases, the signature is allowed to differ.
    """
    lookup_table = format_to_helper[format_type]

    def builder(func):
        if name in lookup_table:
            raise ValueError("Format name '{}' is used twice".format(name))
        lookup_table[name] = func
        if help_msg is not None:
            format_to_help_message[format_type][name] = help_msg
        return func
    return builder


def get_helper_for_format(name, format_type):
    helper = format_to_helper[format_type].get(name)
    if name == "help":
        full_help_message = '\nDetailed help messages:\n\n' + '\n\n'.join(
            '[{}] {}'.format(*x)
            for x in sorted(
                format_to_help_message[format_type].items(),
                key=lambda x: x[0],
            )
        )
        print(
            'List of all formats: {}'.format(
                ', '.join(format_to_helper[format_type].keys())
            ),
            full_help_message if format_to_help_message[format_type] else '',
            file=sys.stderr
        )
        sys.exit(os.EX_USAGE)
    if helper is None:
        sys.stderr.write("Invalid {} format: {}\n".format(format_type, name))
        sys.exit(os.EX_USAGE)
    return helper


@add_format("python", "output")
def __python_output_handler(ret):
    if isinstance(ret, string_types):
        print(ret)
    else:
        pprint.pprint(ret, indent=2)


def __thrift_to_json(x):
    trans = TTransport.TMemoryBuffer()
    proto = TSimpleJSONProtocol.TSimpleJSONProtocol(trans)
    x.write(proto)
    return json.loads(trans.getvalue())


@add_format("json", "output")
def __json_output_handler(ret):
    """
    Python object
    {
        "foo": [
            ThriftStructB(
                x=2
            ),
        ],
        "x": ["%set is nice", 9,8,7, set("blah % blah", 4, 5, 6)],
        "bar": ThriftStructA(
            x=1,
            y="b",
            z=[1,2,3]
        ),
    }

    <=>
    JSON object
    {
        "foo": [
            {"x": 2}
        ],
        "x": ["%set is nice", 9,8,7, ["blah % blah", 4, 5, 6]],
        "bar": {
            "x": 1,
            "y": "b",
            "z": [1,2,3]
        }
    }

    There is no need to handle the type ambiguity between Json dict and
        thrift structs, because pyremote knows what type the services want,
        and we simply try to convert them to that type.

    Also, the exact form of dictionaries produced for Thrift structs may differ
        based across different Thrift versions.
    """
    print(json.dumps(ret, default=__thrift_to_json))


def __eval_arg(arg, thrift_types):
    """Evaluate a commandline argument within the scope of the IF types"""
    code_globals = {}
    code_globals.update(thrift_types)
    # Explicitly compile the code so that it does not inherit our
    # __future__ directives imported above.  In particular this ensures
    # that string literals are not treated as unicode unless explicitly
    # qualified as such.
    code = compile(arg, '<command_line>', 'eval', 0, 1)
    return eval(code, code_globals)


def __preprocess_input(fn, args, ctx):
    if len(args) != len(fn.args):
        sys.stderr.write(('"%s" expects %d arguments (received %d)\n')
                         % (fn.fn_name, len(fn.args), len(args)))
        sys.exit(os.EX_USAGE)

    # Get all custom Thrift types
    return {
        key: getattr(ctx.ttypes, key)
        for key in dir(ctx.ttypes)
    }


@add_format("python", "input", (
    'Evaluate every string in "function_args" using eval() so '
    'that we can support any type of data, unless we already know '
    'the thrift function accepts that argument as a string. In that '
    'case, we simply pass your string without eval().'
))
def __python_natural_input_handler(fn, args, ctx):
    return __python_eval_input_handler(fn, [
        repr(x) if y[2] == 'string' else x
        for x, y in zip(args, fn.args)
    ], ctx)


@add_format("python_eval", "input", (
    'Similar to "python", but we evaluate everything, including strings.'
))
def __python_eval_input_handler(fn, args, ctx):
    thrift_types = __preprocess_input(fn, args, ctx)
    fn_args = []
    for arg in args:
        try:
            value = __eval_arg(arg, thrift_types)
        except Exception:
            traceback.print_exc(file=sys.stderr)
            sys.stderr.write('error parsing argument "%s"' % (arg,))
            sys.exit(os.EX_DATAERR)
        fn_args.append(value)

    return fn_args


@add_format("python_eval_stdin", "input", (
    'Disables the command line option "function_args", and requires '
    'you to pass parameters from stdin. The string you passed in will '
    'be sent to eval(). And it must produce a Python list of objects, '
    'which represents the input argument list to the thrift function.'
))
def __python_stdin_input_handler(fn, args, ctx):
    new_args = json.load(sys.stdin)
    return __python_eval_input_handler(fn, new_args, ctx)


def __args_class_for_function(fn, service_class):
    args_class = getattr(service_class, fn.fn_name + "_args", None)
    if not args_class:
        sys.stderr.write(
            "ERROR: <function name>_args class is unexpected missing. Thrift "
            "may have deprecated its usage. Please re-implement pyremote."
        )
        sys.exit(os.EX_USAGE)
    return args_class


@add_format("json", "input", (
    'Please pass in only one string as "function_args". This string '
    'is a json. Its top level must be a dictionary mapping names of '
    'the thrift function\'s parameters to the value you want to pass '
    'in. Make sure to represent thrift objects using the same format '
    'as generated by pyremote (when using json output format). [Hint: '
    'use this option with a command line tool that can operate on JSONs]'
))
def __json_natural_input_handler(fn, args, ctx):
    if len(args) != 1:
        sys.stderr.write(
            'Error: when using "json" input format, only one cmdline argument '
            'should be used to specify function call arguments. Store arguments '
            'as a json list.'
        )
        sys.exit(os.EX_USAGE)
    partially_decoded = json.loads(args[0])
    if not isinstance(partially_decoded, dict):
        sys.stderr.write(
            "ERROR: Your json input must be a dictionary (of function arguments).\n"
        )
        sys.exit(os.EX_USAGE)
    args_class = __args_class_for_function(fn, ctx.service_class)
    args_obj = args_class()
    args_obj.readFromJson(partially_decoded, is_text=False)
    ans = [getattr(args_obj, arg_name, None) for _, arg_name, _ in fn.args]
    if None in ans:
        sys.stderr.write(
            "ERROR: <function name>_args class is unexpected missing. Thrift "
            "may have deprecated its usage. Please re-implement pyremote."
        )
        sys.exit(os.EX_USAGE)
    return ans


@add_format("json_stdin", "input", (
    'Similar to "json". But this disables the command line option "function_args" '
    'and accepts one json string from stdin.'
))
def __json_stdin_input_handler(fn, args, ctx):
    return __json_natural_input_handler(fn, [sys.stdin.read()], ctx)


def __is_thrift_struct(obj):
    try:
        json.dumps(obj)
        return False
    except BaseException:
        return True


def __get_template_for_struct(struct_type):
    fields = [
        (x[1], x[2], x[3]) for x in
        struct_type.thrift_spec
        if x is not None
    ]
    ans = {}
    for type1, name, type2 in fields:
        if type1 != Thrift.TType.STRUCT:
            ans[name] = "TEMPLATE [TYPE UNKNOWN]"
            continue
        ans[name] = __get_template_for_struct(type2[0])
    return ans


def get_json_template_obj(name, functions, service_class):
    fn = functions.get(name)
    struct = getattr(service_class, name, None)
    if fn is None and struct is None:
        sys.stderr.write("ERROR: unknown structure/function: {}\n".format(name))
        sys.exit(os.EX_USAGE)
    if fn is not None:
        print(
            "Treating", name,
            "as a function. Generating template for its arguments...",
            file=sys.stderr,
        )
        ans_class = __args_class_for_function(fn, service_class)
    elif struct is not None:
        print(
            "Treating", name,
            "as a structure. Generating template for it...",
            file=sys.stderr,
        )
        ans_class = struct
    return __get_template_for_struct(ans_class)


class RemoteClient(object):
    def __init__(self, functions, service_names, service_class,
                 ttypes, print_usage, default_port):
        self.functions = functions
        self.service_names = service_names
        self.service_class = service_class
        self.ttypes = ttypes
        self.print_usage = print_usage
        self.default_port = default_port

    def _exit(self, error_message=None, status=os.EX_USAGE, err_out=sys.stderr):
        """ Report an error, show help information, and exit the program """
        if error_message is not None:
            print("Error: %s" % error_message, file=err_out)

        if status is os.EX_USAGE:
            self.print_usage(err_out)

        if (self.functions is not None and
                status in {os.EX_USAGE, os.EX_CONFIG}):
            print_functions(self.functions, self.service_names, err_out)

        sys.exit(status)

    def _validate_options(self, options):
        """Check option validity and call _exit if there is an error"""
        pass

    def _get_client(self, options):
        """Get the thrift client that will be used to make method calls"""
        raise TypeError("_get_client should be called on "
                        "a subclass of RemoteClient")

    def _close_client(self):
        """After making the method call, do any cleanup work"""
        pass

    def _process_args(self, cmdline_args):
        """Populate instance data using commandline arguments"""
        fn_name = cmdline_args.function_name
        if fn_name not in self.functions:
            self._exit(error_message='Unknown function "%s"' % fn_name,
                       status=os.EX_CONFIG)
        else:
            function = self.functions[fn_name]

        function_args = cmdline_args.input_format(
            function, cmdline_args.function_args, self
        )

        self._validate_options(cmdline_args)
        return function.fn_name, function_args

    def _execute(self, fn_name, fn_args, cmdline_args):
        """Make the requested call.
        Assumes _parse_args() and _process_args() have already been called.
        """
        client = self._get_client(cmdline_args)

        # Call the function
        method = getattr(client, fn_name)
        try:
            ret = method(*fn_args)
        except Thrift.TException as e:
            ret = 'Exception:\n' + str(e)

        cmdline_args.output_format(ret)

        transport = client._iprot.trans
        if isinstance(transport, THeaderTransport):
            response_headers = transport.get_headers()
            if response_headers is not None and len(response_headers) > 0:
                print("Response headers:")
                pprint.pprint(transport.get_headers(), indent=2)

        self._close_client()

    def run(self, cmdline_args):
        fn_name, fn_args = self._process_args(cmdline_args)
        self._execute(fn_name, fn_args, cmdline_args)
        self._exit(status=0)


def ssl_parsed_bool(arg):
    if isinstance(arg, bool):
        return arg
    if arg in ('true', '1'):
        return True
    elif arg in ('false', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('argument must be one of true, 1, false, or 0')

class RemoteTransportClient(RemoteClient):
    """Abstract class for clients with transport manually opened and closed"""
    CMDLINE_OPTIONS = [
        (
            ('-f', '--framed'),
            {
                'action': 'store_true',
                'default': False,
                'help': 'Use framed transport'
            }
        ), (
            ('-s', '--ssl'),
            {
                'action': 'store',
                'type': ssl_parsed_bool,
                'default': True,
                'const': True,
                'nargs': '?',
                'help': 'Use SSL socket'
            }
        ), (
            ('-U', '--unframed'),
            {
                'action': 'store_true',
                'default': False,
                'help': 'Use unframed transport'
            }
        ), (
            ('-j', '--json'),
            {
                'action': 'store_true',
                'default': False,
                'help': 'Use TJSONProtocol'
            }
        ), (
            ('-c', '--compact'),
            {
                'action': 'store_true',
                'default': False,
                'help': 'Use TCompactProtocol'
            }
        ), (
            ('-H', '--headers'),
            {
                'action': 'store',
                'metavar': 'HEADERS_DICT',
                'help':
                'Python code to eval() into a dict of write headers',
            }
        ),
    ]

    def _get_client_by_transport(self, options, transport, socket=None):
        # Create the protocol and client
        if options.json:
            protocol = TJSONProtocol.TJSONProtocol(transport)
        elif options.compact:
            protocol = TCompactProtocol.TCompactProtocol(transport)

        # No explicit option about protocol is specified. Try to infer.
        elif options.framed or options.unframed:
            protocol = TBinaryProtocol.TBinaryProtocolAccelerated(transport)

        elif socket is not None:
            # If json, compact, framed, and unframed are not specified,
            # THeaderProtocol is the default.
            transport = THeaderTransport(socket)
            if options.headers is not None:
                try:
                    parsed_headers = eval(options.headers)
                except Exception:
                    self._exit(
                        error_message='Request headers (--headers) argument'
                                      ' failed eval')
                if not isinstance(parsed_headers, dict):
                    self._exit(
                        error_message='Request headers (--headers) argument'
                                      ' must evaluate to a dict')
                for header_name, header_value in parsed_headers.items():
                    transport.set_header(header_name, header_value)
            protocol = THeaderProtocol.THeaderProtocol(transport)
        else:
            self._exit(error_message=('No valid protocol '
                                      'specified for %s' % (type(self))),
                       status=os.EX_USAGE)

        transport.open()
        self._transport = transport

        client = self.service_class.Client(protocol)

        return client

    def _close_client(self):
        self._transport.close()

    def _validate_options(self, options):
        super(RemoteTransportClient, self)._validate_options(options)
        if options.framed and options.unframed:
            self._exit(error_message='cannot specify both '
                       '--framed and --unframed')

    def _parse_host_port(self, value, default_port):
        parts = value.rsplit(':', 1)
        if len(parts) == 1:
            return (parts[0], default_port)
        try:
            port = int(parts[1])
        except ValueError:
            raise ValueError('invalid port: ' + parts[1])
        return (parts[0], port)


class RemoteHostClient(RemoteTransportClient):
    SELECTOR_OPTIONS = 'host'
    CMDLINE_OPTIONS = list(RemoteTransportClient.CMDLINE_OPTIONS) + [(
        ('-h', '--host'),
        {
            'action': 'store',
            'metavar': 'HOST[:PORT]',
            'help': 'The host and port to connect to'
        }
    )]

    def _get_client(self, options):
        host, port = self._parse_host_port(options.host, self.default_port)
        socket = (TSSLSocket.TSSLSocket(host, port) if options.ssl
                  else TSocket.TSocket(host, port))
        if options.framed:
            transport = TTransport.TFramedTransport(socket)
        else:
            transport = TTransport.TBufferedTransport(socket)
        return self._get_client_by_transport(options, transport, socket=socket)


class RemoteHttpClient(RemoteTransportClient):
    SELECTOR_OPTIONS = 'url'
    CMDLINE_OPTIONS = list(RemoteTransportClient.CMDLINE_OPTIONS) + [(
        ('-u', '--url'),
        {
            'action': 'store',
            'help': 'The URL to connect to, for HTTP transport'
        }
    )]

    def _get_client(self, options):
        url = urlparse(options.url)
        host, port = self._parse_host_port(url[1], 80)
        transport = THttpClient.THttpClient(options.url)
        return self._get_client_by_transport(options, transport)

    def _validate_options(self, options):
        """Check if there are any option inconsistencies, and exit if so"""
        super(RemoteHttpClient, self)._validate_options(options)
        if not any([options.unframed, options.json]):
            self._exit(error_message='can only specify --url with '
                       '--unframed or --json')


class RemoteUNIXDomainClient(RemoteTransportClient):
    SELECTOR_OPTIONS = 'path'
    CMDLINE_OPTIONS = list(RemoteTransportClient.CMDLINE_OPTIONS) + [(
        ('-p', '--path'),
        {
            'action': 'store',
            'help': 'The path of the socket to use'
        }
    )]

    def _get_client(self, options):
        socket = TSocket.TSocket(unix_socket=options.path)
        if options.framed:
            transport = TTransport.TFramedTransport(socket)
        else:
            transport = TTransport.TBufferedTransport(socket)
        return self._get_client_by_transport(options, transport, socket=socket)


class Namespace(object):
    def __init__(self, attrs=None):
        if attrs is not None:
            self.__dict__.update(attrs)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        self.__dict__[key] = value


class Remote(object):
    __client_types = set()
    __occupied_args = {}
    __parser = argparse.ArgumentParser(add_help=False)

    @classmethod
    def register_cmdline_options(cls, cmdline_options):
        for args, kwargs in cmdline_options:
            is_repeated = False
            for arg in args:
                if arg in cls.__occupied_args:
                    if cls.__occupied_args[arg] != kwargs:
                        raise ValueError('Redefinition of {}'.format(arg))
                    is_repeated = True
            if is_repeated:
                continue
            cls.__occupied_args.update({x: kwargs for x in args})
            cls.__parser.add_argument(*args, **kwargs)

    @classmethod
    def register_client_type(cls, client_type):
        if not issubclass(client_type, RemoteClient):
            raise TypeError(('Remote client must be of type RemoteClient. '
                             'Got type %s.' % client_type.__name__))
        if client_type is RemoteClient:
            raise TypeError(('Remote client must be a strict subclass '
                             'of RemoteClient.'))
        if not hasattr(client_type, 'SELECTOR_OPTIONS'):
            raise AttributeError(('Remote client must have a '
                                  'SELECTOR_OPTIONS field.'))

        cls.__client_types.add(client_type)
        cls.register_cmdline_options(client_type.CMDLINE_OPTIONS)

    @classmethod
    def _exit_usage_error(cls, message):
        sys.stderr.write('ERROR: ' + message + '\n')
        cls.__parser.print_help(sys.stderr)
        sys.exit(os.EX_USAGE)

    @classmethod
    def _get_client_type(cls, options):
        matching_types = [ct for ct in cls.__client_types if
                          getattr(options, ct.SELECTOR_OPTIONS) is not None]
        if len(matching_types) != 1:
            cls._exit_usage_error('Must specify exactly one of [%s]' % (
                ', '.join('--%s' % ct.SELECTOR_OPTIONS
                          for ct in cls.__client_types)))
        else:
            return matching_types[0]

    @classmethod
    def _parse_cmdline_options(cls, argv):
        cls.register_cmdline_options((
            (
                ('-ifmt', '--input-format'),
                {
                    'action': 'store',
                    'default': 'python',
                    'type': lambda x: get_helper_for_format(x, "input"),
                    'help': (
                        'Change the format for function_args. Generally speaking, '
                        'there are two main formats: python_* and json_*. Defaults '
                        'to "python". Use -ifmt help for entire list of available '
                        'formats.'
                    ),
                },
            ),
            (
                ('-ofmt', '--output-format', ),
                {
                    'action': 'store',
                    'default': 'python',
                    'type': lambda x: get_helper_for_format(x, "output"),
                    'help': (
                        'Change the output format for the return value. The '
                        'default is "python", which direclty prints out strings '
                        'and pprint() other types. Available formats: {}.'
                    ).format(','.join(format_to_helper["output"].keys()))
                },
            ),
            (
                ('--help', ),
                {'action': 'help'},
            ),
            (
                ('-la', '--list-all-functions'),
                {'action': 'store_true'},
            ),
            (
                ('-l', '--list-functions', ),
                {'action': 'store_true'},
            ),
            (
                ('-g', '--generate-template'),
                {
                    'action': 'store',
                    'metavar': 'THRIFT_STRUCT_OR_FUNCTION_NAME',
                    'help': (
                        'Generate a template for a thrift struct, OR, arguments of '
                        'a function call. Currently it supports only json format. '
                        'No need to specify function_name.'
                    )
                }
            ),
            (
                ('function_name', ),
                {'nargs': '?', 'help': 'Name of the remote function to call'},
            ),
            (
                ('function_args', ),
                {'nargs': '*', 'help': (
                    'Arguments for the remote function. Look at --input-format '
                    'for more details.'
                )},
            ),
        ))
        try:
            return cls.__parser.parse_args(argv[1:])
        except BaseException:
            sys.exit(os.EX_USAGE)

    @classmethod
    def run(cls, functions, service_names, service_class,
            ttypes, argv, default_port=9090):
        args = cls._parse_cmdline_options(argv)
        conflicts = [x for x in [
            "list_all_functions",
            "list_functions",
            "generate_template",
        ] if getattr(args, x)]
        if len(conflicts) > 1:
            cls._exit_usage_error(
                'Please do not specify all of {} at once.'.format(
                    ','.join(conflicts)
                )
            )
        if args.list_all_functions:
            print_functions(functions, service_names, sys.stdout, local_only=False)
            return
        if args.list_functions:
            print_functions(functions, service_names, sys.stdout, local_only=True)
            return
        if args.function_name is None:
            cls._exit_usage_error('Please specify function_name.')
        if args.generate_template:
            ans = get_json_template_obj(
                args.generate_template, functions, service_class
            )
            print(json.dumps(ans))
            return
        client_type = cls._get_client_type(args)
        client = client_type(functions, service_names, service_class, ttypes,
                             cls.__parser.print_help, default_port)
        client.run(args)


Remote.register_client_type(RemoteHostClient)
Remote.register_client_type(RemoteHttpClient)
Remote.register_client_type(RemoteUNIXDomainClient)



#!/usr/bin/env python
# --coding:utf-8--

# Copyright (c) 2020 vesoft inc. All rights reserved.
#
# This source code is licensed under Apache 2.0 License,
# attached with Common Clause Condition 1.0, found in the LICENSES directory.

from nebula2.common.ttypes import Value
Value.__hash__ = lambda self: self.value.__hash__()





#
# Autogenerated by Thrift
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#  @generated
#

from __future__ import absolute_import
import six
import sys
from nebula2.fbthrift.util.Recursive import fix_spec
from nebula2.fbthrift.Thrift import TType, TMessageType, TPriority, TRequestContext, TProcessorEventHandler, TServerInterface, TProcessor, TException, TApplicationException, UnimplementedTypedef
from nebula2.fbthrift.protocol.TProtocol import TProtocolException


from .ttypes import UTF8STRINGS, ProfilingStats, PlanNodeBranchInfo, Pair, PlanNodeDescription, PlanDescription, ExecutionResponse, AuthResponse, VerifyClientVersionResp, VerifyClientVersionReq
import nebula2.common.ttypes

from nebula2.fbthrift.Thrift import TProcessor
import pprint
import warnings
from nebula2.fbthrift import Thrift
from nebula2.fbthrift.transport import TTransport
from nebula2.fbthrift.protocol import TBinaryProtocol
from nebula2.fbthrift.protocol import TCompactProtocol
from nebula2.fbthrift.protocol import THeaderProtocol
fastproto = None
try:
  from nebula2.fbthrift.protocol import fastproto
except ImportError:
  pass

all_structs = []
UTF8STRINGS = bool(0) or sys.version_info.major >= 3

from nebula2.fbthrift.util.Decorators import (
  future_process_main,
  future_process_method,
  process_main as thrift_process_main,
  process_method as thrift_process_method,
  should_run_on_thread,
  write_results_after_future,
)

class Iface:
  def authenticate(self, username=None, password=None):
    """
    Parameters:
     - username
     - password
    """
    pass

  def signout(self, sessionId=None):
    """
    Parameters:
     - sessionId
    """
    pass

  def execute(self, sessionId=None, stmt=None):
    """
    Parameters:
     - sessionId
     - stmt
    """
    pass

  def executeJson(self, sessionId=None, stmt=None):
    """
    Parameters:
     - sessionId
     - stmt
    """
    pass

  def verifyClientVersion(self, req=None):
    """
    Parameters:
     - req
    """
    pass


class ContextIface:
  def authenticate(self, handler_ctx, username=None, password=None):
    """
    Parameters:
     - username
     - password
    """
    pass

  def signout(self, handler_ctx, sessionId=None):
    """
    Parameters:
     - sessionId
    """
    pass

  def execute(self, handler_ctx, sessionId=None, stmt=None):
    """
    Parameters:
     - sessionId
     - stmt
    """
    pass

  def executeJson(self, handler_ctx, sessionId=None, stmt=None):
    """
    Parameters:
     - sessionId
     - stmt
    """
    pass

  def verifyClientVersion(self, handler_ctx, req=None):
    """
    Parameters:
     - req
    """
    pass


# HELPER FUNCTIONS AND STRUCTURES

class authenticate_args:
  """
  Attributes:
   - username
   - password
  """

  thrift_spec = None
  thrift_field_annotations = None
  thrift_struct_annotations = None
  __init__ = None
  @staticmethod
  def isUnion():
    return False

  def read(self, iprot):
    if (isinstance(iprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0)
      return
    if (isinstance(iprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2)
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.username = iprot.readString()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.password = iprot.readString()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if (isinstance(oprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0))
      return
    if (isinstance(oprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2))
      return
    oprot.writeStructBegin('authenticate_args')
    if self.username != None:
      oprot.writeFieldBegin('username', TType.STRING, 1)
      oprot.writeString(self.username)
      oprot.writeFieldEnd()
    if self.password != None:
      oprot.writeFieldBegin('password', TType.STRING, 2)
      oprot.writeString(self.password)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def __repr__(self):
    L = []
    padding = ' ' * 4
    if self.username is not None:
      value = pprint.pformat(self.username, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    username=%s' % (value))
    if self.password is not None:
      value = pprint.pformat(self.password, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    password=%s' % (value))
    return "%s(%s)" % (self.__class__.__name__, "\n" + ",\n".join(L) if L else '')

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return False

    return self.__dict__ == other.__dict__ 

  def __ne__(self, other):
    return not (self == other)

  # Override the __hash__ function for Python3 - t10434117
  if not six.PY2:
    __hash__ = object.__hash__

all_structs.append(authenticate_args)
authenticate_args.thrift_spec = (
  None, # 0
  (1, TType.STRING, 'username', False, None, 2, ), # 1
  (2, TType.STRING, 'password', False, None, 2, ), # 2
)

authenticate_args.thrift_struct_annotations = {
}
authenticate_args.thrift_field_annotations = {
}

def authenticate_args__init__(self, username=None, password=None,):
  self.username = username
  self.password = password

authenticate_args.__init__ = authenticate_args__init__

def authenticate_args__setstate__(self, state):
  state.setdefault('username', None)
  state.setdefault('password', None)
  self.__dict__ = state

authenticate_args.__getstate__ = lambda self: self.__dict__.copy()
authenticate_args.__setstate__ = authenticate_args__setstate__

class authenticate_result:
  """
  Attributes:
   - success
  """

  thrift_spec = None
  thrift_field_annotations = None
  thrift_struct_annotations = None
  __init__ = None
  @staticmethod
  def isUnion():
    return False

  def read(self, iprot):
    if (isinstance(iprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0)
      return
    if (isinstance(iprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2)
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = AuthResponse()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if (isinstance(oprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0))
      return
    if (isinstance(oprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2))
      return
    oprot.writeStructBegin('authenticate_result')
    if self.success != None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def __repr__(self):
    L = []
    padding = ' ' * 4
    if self.success is not None:
      value = pprint.pformat(self.success, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    success=%s' % (value))
    return "%s(%s)" % (self.__class__.__name__, "\n" + ",\n".join(L) if L else '')

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return False

    return self.__dict__ == other.__dict__ 

  def __ne__(self, other):
    return not (self == other)

  # Override the __hash__ function for Python3 - t10434117
  if not six.PY2:
    __hash__ = object.__hash__

all_structs.append(authenticate_result)
authenticate_result.thrift_spec = (
  (0, TType.STRUCT, 'success', [AuthResponse, AuthResponse.thrift_spec, False], None, 2, ), # 0
)

authenticate_result.thrift_struct_annotations = {
}
authenticate_result.thrift_field_annotations = {
}

def authenticate_result__init__(self, success=None,):
  self.success = success

authenticate_result.__init__ = authenticate_result__init__

def authenticate_result__setstate__(self, state):
  state.setdefault('success', None)
  self.__dict__ = state

authenticate_result.__getstate__ = lambda self: self.__dict__.copy()
authenticate_result.__setstate__ = authenticate_result__setstate__

class signout_args:
  """
  Attributes:
   - sessionId
  """

  thrift_spec = None
  thrift_field_annotations = None
  thrift_struct_annotations = None
  __init__ = None
  @staticmethod
  def isUnion():
    return False

  def read(self, iprot):
    if (isinstance(iprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0)
      return
    if (isinstance(iprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2)
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I64:
          self.sessionId = iprot.readI64()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if (isinstance(oprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0))
      return
    if (isinstance(oprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2))
      return
    oprot.writeStructBegin('signout_args')
    if self.sessionId != None:
      oprot.writeFieldBegin('sessionId', TType.I64, 1)
      oprot.writeI64(self.sessionId)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def __repr__(self):
    L = []
    padding = ' ' * 4
    if self.sessionId is not None:
      value = pprint.pformat(self.sessionId, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    sessionId=%s' % (value))
    return "%s(%s)" % (self.__class__.__name__, "\n" + ",\n".join(L) if L else '')

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return False

    return self.__dict__ == other.__dict__ 

  def __ne__(self, other):
    return not (self == other)

  # Override the __hash__ function for Python3 - t10434117
  if not six.PY2:
    __hash__ = object.__hash__

all_structs.append(signout_args)
signout_args.thrift_spec = (
  None, # 0
  (1, TType.I64, 'sessionId', None, None, 2, ), # 1
)

signout_args.thrift_struct_annotations = {
}
signout_args.thrift_field_annotations = {
}

def signout_args__init__(self, sessionId=None,):
  self.sessionId = sessionId

signout_args.__init__ = signout_args__init__

def signout_args__setstate__(self, state):
  state.setdefault('sessionId', None)
  self.__dict__ = state

signout_args.__getstate__ = lambda self: self.__dict__.copy()
signout_args.__setstate__ = signout_args__setstate__

class execute_args:
  """
  Attributes:
   - sessionId
   - stmt
  """

  thrift_spec = None
  thrift_field_annotations = None
  thrift_struct_annotations = None
  __init__ = None
  @staticmethod
  def isUnion():
    return False

  def read(self, iprot):
    if (isinstance(iprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0)
      return
    if (isinstance(iprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2)
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I64:
          self.sessionId = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.stmt = iprot.readString()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if (isinstance(oprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0))
      return
    if (isinstance(oprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2))
      return
    oprot.writeStructBegin('execute_args')
    if self.sessionId != None:
      oprot.writeFieldBegin('sessionId', TType.I64, 1)
      oprot.writeI64(self.sessionId)
      oprot.writeFieldEnd()
    if self.stmt != None:
      oprot.writeFieldBegin('stmt', TType.STRING, 2)
      oprot.writeString(self.stmt)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def __repr__(self):
    L = []
    padding = ' ' * 4
    if self.sessionId is not None:
      value = pprint.pformat(self.sessionId, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    sessionId=%s' % (value))
    if self.stmt is not None:
      value = pprint.pformat(self.stmt, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    stmt=%s' % (value))
    return "%s(%s)" % (self.__class__.__name__, "\n" + ",\n".join(L) if L else '')

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return False

    return self.__dict__ == other.__dict__ 

  def __ne__(self, other):
    return not (self == other)

  # Override the __hash__ function for Python3 - t10434117
  if not six.PY2:
    __hash__ = object.__hash__

all_structs.append(execute_args)
execute_args.thrift_spec = (
  None, # 0
  (1, TType.I64, 'sessionId', None, None, 2, ), # 1
  (2, TType.STRING, 'stmt', False, None, 2, ), # 2
)

execute_args.thrift_struct_annotations = {
}
execute_args.thrift_field_annotations = {
}

def execute_args__init__(self, sessionId=None, stmt=None,):
  self.sessionId = sessionId
  self.stmt = stmt

execute_args.__init__ = execute_args__init__

def execute_args__setstate__(self, state):
  state.setdefault('sessionId', None)
  state.setdefault('stmt', None)
  self.__dict__ = state

execute_args.__getstate__ = lambda self: self.__dict__.copy()
execute_args.__setstate__ = execute_args__setstate__

class execute_result:
  """
  Attributes:
   - success
  """

  thrift_spec = None
  thrift_field_annotations = None
  thrift_struct_annotations = None
  __init__ = None
  @staticmethod
  def isUnion():
    return False

  def read(self, iprot):
    if (isinstance(iprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0)
      return
    if (isinstance(iprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2)
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = ExecutionResponse()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if (isinstance(oprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0))
      return
    if (isinstance(oprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2))
      return
    oprot.writeStructBegin('execute_result')
    if self.success != None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def __repr__(self):
    L = []
    padding = ' ' * 4
    if self.success is not None:
      value = pprint.pformat(self.success, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    success=%s' % (value))
    return "%s(%s)" % (self.__class__.__name__, "\n" + ",\n".join(L) if L else '')

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return False

    return self.__dict__ == other.__dict__ 

  def __ne__(self, other):
    return not (self == other)

  # Override the __hash__ function for Python3 - t10434117
  if not six.PY2:
    __hash__ = object.__hash__

all_structs.append(execute_result)
execute_result.thrift_spec = (
  (0, TType.STRUCT, 'success', [ExecutionResponse, ExecutionResponse.thrift_spec, False], None, 2, ), # 0
)

execute_result.thrift_struct_annotations = {
}
execute_result.thrift_field_annotations = {
}

def execute_result__init__(self, success=None,):
  self.success = success

execute_result.__init__ = execute_result__init__

def execute_result__setstate__(self, state):
  state.setdefault('success', None)
  self.__dict__ = state

execute_result.__getstate__ = lambda self: self.__dict__.copy()
execute_result.__setstate__ = execute_result__setstate__

class executeJson_args:
  """
  Attributes:
   - sessionId
   - stmt
  """

  thrift_spec = None
  thrift_field_annotations = None
  thrift_struct_annotations = None
  __init__ = None
  @staticmethod
  def isUnion():
    return False

  def read(self, iprot):
    if (isinstance(iprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0)
      return
    if (isinstance(iprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2)
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I64:
          self.sessionId = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.stmt = iprot.readString()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if (isinstance(oprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0))
      return
    if (isinstance(oprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2))
      return
    oprot.writeStructBegin('executeJson_args')
    if self.sessionId != None:
      oprot.writeFieldBegin('sessionId', TType.I64, 1)
      oprot.writeI64(self.sessionId)
      oprot.writeFieldEnd()
    if self.stmt != None:
      oprot.writeFieldBegin('stmt', TType.STRING, 2)
      oprot.writeString(self.stmt)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def __repr__(self):
    L = []
    padding = ' ' * 4
    if self.sessionId is not None:
      value = pprint.pformat(self.sessionId, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    sessionId=%s' % (value))
    if self.stmt is not None:
      value = pprint.pformat(self.stmt, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    stmt=%s' % (value))
    return "%s(%s)" % (self.__class__.__name__, "\n" + ",\n".join(L) if L else '')

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return False

    return self.__dict__ == other.__dict__ 

  def __ne__(self, other):
    return not (self == other)

  # Override the __hash__ function for Python3 - t10434117
  if not six.PY2:
    __hash__ = object.__hash__

all_structs.append(executeJson_args)
executeJson_args.thrift_spec = (
  None, # 0
  (1, TType.I64, 'sessionId', None, None, 2, ), # 1
  (2, TType.STRING, 'stmt', False, None, 2, ), # 2
)

executeJson_args.thrift_struct_annotations = {
}
executeJson_args.thrift_field_annotations = {
}

def executeJson_args__init__(self, sessionId=None, stmt=None,):
  self.sessionId = sessionId
  self.stmt = stmt

executeJson_args.__init__ = executeJson_args__init__

def executeJson_args__setstate__(self, state):
  state.setdefault('sessionId', None)
  state.setdefault('stmt', None)
  self.__dict__ = state

executeJson_args.__getstate__ = lambda self: self.__dict__.copy()
executeJson_args.__setstate__ = executeJson_args__setstate__

class executeJson_result:
  """
  Attributes:
   - success
  """

  thrift_spec = None
  thrift_field_annotations = None
  thrift_struct_annotations = None
  __init__ = None
  @staticmethod
  def isUnion():
    return False

  def read(self, iprot):
    if (isinstance(iprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0)
      return
    if (isinstance(iprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2)
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRING:
          self.success = iprot.readString()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if (isinstance(oprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0))
      return
    if (isinstance(oprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2))
      return
    oprot.writeStructBegin('executeJson_result')
    if self.success != None:
      oprot.writeFieldBegin('success', TType.STRING, 0)
      oprot.writeString(self.success)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def __repr__(self):
    L = []
    padding = ' ' * 4
    if self.success is not None:
      value = pprint.pformat(self.success, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    success=%s' % (value))
    return "%s(%s)" % (self.__class__.__name__, "\n" + ",\n".join(L) if L else '')

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return False

    return self.__dict__ == other.__dict__ 

  def __ne__(self, other):
    return not (self == other)

  # Override the __hash__ function for Python3 - t10434117
  if not six.PY2:
    __hash__ = object.__hash__

all_structs.append(executeJson_result)
executeJson_result.thrift_spec = (
  (0, TType.STRING, 'success', False, None, 2, ), # 0
)

executeJson_result.thrift_struct_annotations = {
}
executeJson_result.thrift_field_annotations = {
}

def executeJson_result__init__(self, success=None,):
  self.success = success

executeJson_result.__init__ = executeJson_result__init__

def executeJson_result__setstate__(self, state):
  state.setdefault('success', None)
  self.__dict__ = state

executeJson_result.__getstate__ = lambda self: self.__dict__.copy()
executeJson_result.__setstate__ = executeJson_result__setstate__

class verifyClientVersion_args:
  """
  Attributes:
   - req
  """

  thrift_spec = None
  thrift_field_annotations = None
  thrift_struct_annotations = None
  __init__ = None
  @staticmethod
  def isUnion():
    return False

  def read(self, iprot):
    if (isinstance(iprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0)
      return
    if (isinstance(iprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2)
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.req = VerifyClientVersionReq()
          self.req.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if (isinstance(oprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0))
      return
    if (isinstance(oprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2))
      return
    oprot.writeStructBegin('verifyClientVersion_args')
    if self.req != None:
      oprot.writeFieldBegin('req', TType.STRUCT, 1)
      self.req.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def __repr__(self):
    L = []
    padding = ' ' * 4
    if self.req is not None:
      value = pprint.pformat(self.req, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    req=%s' % (value))
    return "%s(%s)" % (self.__class__.__name__, "\n" + ",\n".join(L) if L else '')

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return False

    return self.__dict__ == other.__dict__ 

  def __ne__(self, other):
    return not (self == other)

  # Override the __hash__ function for Python3 - t10434117
  if not six.PY2:
    __hash__ = object.__hash__

all_structs.append(verifyClientVersion_args)
verifyClientVersion_args.thrift_spec = (
  None, # 0
  (1, TType.STRUCT, 'req', [VerifyClientVersionReq, VerifyClientVersionReq.thrift_spec, False], None, 2, ), # 1
)

verifyClientVersion_args.thrift_struct_annotations = {
}
verifyClientVersion_args.thrift_field_annotations = {
}

def verifyClientVersion_args__init__(self, req=None,):
  self.req = req

verifyClientVersion_args.__init__ = verifyClientVersion_args__init__

def verifyClientVersion_args__setstate__(self, state):
  state.setdefault('req', None)
  self.__dict__ = state

verifyClientVersion_args.__getstate__ = lambda self: self.__dict__.copy()
verifyClientVersion_args.__setstate__ = verifyClientVersion_args__setstate__

class verifyClientVersion_result:
  """
  Attributes:
   - success
  """

  thrift_spec = None
  thrift_field_annotations = None
  thrift_struct_annotations = None
  __init__ = None
  @staticmethod
  def isUnion():
    return False

  def read(self, iprot):
    if (isinstance(iprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0)
      return
    if (isinstance(iprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(iprot, THeaderProtocol.THeaderProtocolAccelerate) and iprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastproto is not None:
      fastproto.decode(self, iprot.trans, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2)
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = VerifyClientVersionResp()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if (isinstance(oprot, TBinaryProtocol.TBinaryProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_BINARY_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=0))
      return
    if (isinstance(oprot, TCompactProtocol.TCompactProtocolAccelerated) or (isinstance(oprot, THeaderProtocol.THeaderProtocolAccelerate) and oprot.get_protocol_id() == THeaderProtocol.THeaderProtocol.T_COMPACT_PROTOCOL)) and self.thrift_spec is not None and fastproto is not None:
      oprot.trans.write(fastproto.encode(self, [self.__class__, self.thrift_spec, False], utf8strings=UTF8STRINGS, protoid=2))
      return
    oprot.writeStructBegin('verifyClientVersion_result')
    if self.success != None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def __repr__(self):
    L = []
    padding = ' ' * 4
    if self.success is not None:
      value = pprint.pformat(self.success, indent=0)
      value = padding.join(value.splitlines(True))
      L.append('    success=%s' % (value))
    return "%s(%s)" % (self.__class__.__name__, "\n" + ",\n".join(L) if L else '')

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return False

    return self.__dict__ == other.__dict__ 

  def __ne__(self, other):
    return not (self == other)

  # Override the __hash__ function for Python3 - t10434117
  if not six.PY2:
    __hash__ = object.__hash__

all_structs.append(verifyClientVersion_result)
verifyClientVersion_result.thrift_spec = (
  (0, TType.STRUCT, 'success', [VerifyClientVersionResp, VerifyClientVersionResp.thrift_spec, False], None, 2, ), # 0
)

verifyClientVersion_result.thrift_struct_annotations = {
}
verifyClientVersion_result.thrift_field_annotations = {
}

def verifyClientVersion_result__init__(self, success=None,):
  self.success = success

verifyClientVersion_result.__init__ = verifyClientVersion_result__init__

def verifyClientVersion_result__setstate__(self, state):
  state.setdefault('success', None)
  self.__dict__ = state

verifyClientVersion_result.__getstate__ = lambda self: self.__dict__.copy()
verifyClientVersion_result.__setstate__ = verifyClientVersion_result__setstate__

class Client(Iface):
  def __enter__(self):
    return self

  def __exit__(self, type, value, tb):
    self._iprot.trans.close()
    if self._iprot is not self._oprot:
      self._oprot.trans.close()

  def __init__(self, iprot, oprot=None):
    self._iprot = self._oprot = iprot
    if oprot != None:
      self._oprot = oprot
    self._seqid = 0

  def authenticate(self, username=None, password=None):
    """
    Parameters:
     - username
     - password
    """
    self.send_authenticate(username, password)
    return self.recv_authenticate()

  def send_authenticate(self, username=None, password=None):
    self._oprot.writeMessageBegin('authenticate', TMessageType.CALL, self._seqid)
    args = authenticate_args()
    args.username = username
    args.password = password
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_authenticate(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = authenticate_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    if result.success != None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "authenticate failed: unknown result");

  def signout(self, sessionId=None):
    """
    Parameters:
     - sessionId
    """
    self.send_signout(sessionId)

  def send_signout(self, sessionId=None):
    self._oprot.writeMessageBegin('signout', TMessageType.CALL, self._seqid)
    args = signout_args()
    args.sessionId = sessionId
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.onewayFlush()
  def execute(self, sessionId=None, stmt=None):
    """
    Parameters:
     - sessionId
     - stmt
    """
    self.send_execute(sessionId, stmt)
    return self.recv_execute()

  def send_execute(self, sessionId=None, stmt=None):
    self._oprot.writeMessageBegin('execute', TMessageType.CALL, self._seqid)
    args = execute_args()
    args.sessionId = sessionId
    args.stmt = stmt
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_execute(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = execute_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    if result.success != None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "execute failed: unknown result");

  def executeJson(self, sessionId=None, stmt=None):
    """
    Parameters:
     - sessionId
     - stmt
    """
    self.send_executeJson(sessionId, stmt)
    return self.recv_executeJson()

  def send_executeJson(self, sessionId=None, stmt=None):
    self._oprot.writeMessageBegin('executeJson', TMessageType.CALL, self._seqid)
    args = executeJson_args()
    args.sessionId = sessionId
    args.stmt = stmt
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_executeJson(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = executeJson_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    if result.success != None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "executeJson failed: unknown result");

  def verifyClientVersion(self, req=None):
    """
    Parameters:
     - req
    """
    self.send_verifyClientVersion(req)
    return self.recv_verifyClientVersion()

  def send_verifyClientVersion(self, req=None):
    self._oprot.writeMessageBegin('verifyClientVersion', TMessageType.CALL, self._seqid)
    args = verifyClientVersion_args()
    args.req = req
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_verifyClientVersion(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = verifyClientVersion_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    if result.success != None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "verifyClientVersion failed: unknown result");


class Processor(Iface, TProcessor):
  _onewayMethods = ("signout",)

  def __init__(self, handler):
    TProcessor.__init__(self)
    self._handler = handler
    self._processMap = {}
    self._priorityMap = {}
    self._processMap["authenticate"] = Processor.process_authenticate
    self._priorityMap["authenticate"] = TPriority.NORMAL
    self._processMap["signout"] = Processor.process_signout
    self._priorityMap["signout"] = TPriority.NORMAL
    self._processMap["execute"] = Processor.process_execute
    self._priorityMap["execute"] = TPriority.NORMAL
    self._processMap["executeJson"] = Processor.process_executeJson
    self._priorityMap["executeJson"] = TPriority.NORMAL
    self._processMap["verifyClientVersion"] = Processor.process_verifyClientVersion
    self._priorityMap["verifyClientVersion"] = TPriority.NORMAL

  def onewayMethods(self):
    l = []
    l.extend(Processor._onewayMethods)
    return tuple(l)

  @thrift_process_main()
  def process(self,): pass

  @thrift_process_method(authenticate_args, oneway=False)
  def process_authenticate(self, args, handler_ctx):
    result = authenticate_result()
    try:
      result.success = self._handler.authenticate(args.username, args.password)
    except:
      ex = sys.exc_info()[1]
      self._event_handler.handlerError(handler_ctx, 'authenticate', ex)
      result = Thrift.TApplicationException(message=repr(ex))
    return result

  @thrift_process_method(signout_args, oneway=True)
  def process_signout(self, args, handler_ctx):
    try:
      self._handler.signout(args.sessionId)
    except:
      ex = sys.exc_info()[1]
      self._event_handler.handlerError(handler_ctx, 'signout', ex)
      result = Thrift.TApplicationException(message=repr(ex))

  @thrift_process_method(execute_args, oneway=False)
  def process_execute(self, args, handler_ctx):
    result = execute_result()
    try:
      result.success = self._handler.execute(args.sessionId, args.stmt)
    except:
      ex = sys.exc_info()[1]
      self._event_handler.handlerError(handler_ctx, 'execute', ex)
      result = Thrift.TApplicationException(message=repr(ex))
    return result

  @thrift_process_method(executeJson_args, oneway=False)
  def process_executeJson(self, args, handler_ctx):
    result = executeJson_result()
    try:
      result.success = self._handler.executeJson(args.sessionId, args.stmt)
    except:
      ex = sys.exc_info()[1]
      self._event_handler.handlerError(handler_ctx, 'executeJson', ex)
      result = Thrift.TApplicationException(message=repr(ex))
    return result

  @thrift_process_method(verifyClientVersion_args, oneway=False)
  def process_verifyClientVersion(self, args, handler_ctx):
    result = verifyClientVersion_result()
    try:
      result.success = self._handler.verifyClientVersion(args.req)
    except:
      ex = sys.exc_info()[1]
      self._event_handler.handlerError(handler_ctx, 'verifyClientVersion', ex)
      result = Thrift.TApplicationException(message=repr(ex))
    return result

Iface._processor_type = Processor

class ContextProcessor(ContextIface, TProcessor):
  _onewayMethods = ("signout",)

  def __init__(self, handler):
    TProcessor.__init__(self)
    self._handler = handler
    self._processMap = {}
    self._priorityMap = {}
    self._processMap["authenticate"] = ContextProcessor.process_authenticate
    self._priorityMap["authenticate"] = TPriority.NORMAL
    self._processMap["signout"] = ContextProcessor.process_signout
    self._priorityMap["signout"] = TPriority.NORMAL
    self._processMap["execute"] = ContextProcessor.process_execute
    self._priorityMap["execute"] = TPriority.NORMAL
    self._processMap["executeJson"] = ContextProcessor.process_executeJson
    self._priorityMap["executeJson"] = TPriority.NORMAL
    self._processMap["verifyClientVersion"] = ContextProcessor.process_verifyClientVersion
    self._priorityMap["verifyClientVersion"] = TPriority.NORMAL

  def onewayMethods(self):
    l = []
    l.extend(ContextProcessor._onewayMethods)
    return tuple(l)

  @thrift_process_main()
  def process(self,): pass

  @thrift_process_method(authenticate_args, oneway=False)
  def process_authenticate(self, args, handler_ctx):
    result = authenticate_result()
    try:
      result.success = self._handler.authenticate(handler_ctx, args.username, args.password)
    except:
      ex = sys.exc_info()[1]
      self._event_handler.handlerError(handler_ctx, 'authenticate', ex)
      result = Thrift.TApplicationException(message=repr(ex))
    return result

  @thrift_process_method(signout_args, oneway=True)
  def process_signout(self, args, handler_ctx):
    try:
      self._handler.signout(handler_ctx, args.sessionId)
    except:
      ex = sys.exc_info()[1]
      self._event_handler.handlerError(handler_ctx, 'signout', ex)
      result = Thrift.TApplicationException(message=repr(ex))

  @thrift_process_method(execute_args, oneway=False)
  def process_execute(self, args, handler_ctx):
    result = execute_result()
    try:
      result.success = self._handler.execute(handler_ctx, args.sessionId, args.stmt)
    except:
      ex = sys.exc_info()[1]
      self._event_handler.handlerError(handler_ctx, 'execute', ex)
      result = Thrift.TApplicationException(message=repr(ex))
    return result

  @thrift_process_method(executeJson_args, oneway=False)
  def process_executeJson(self, args, handler_ctx):
    result = executeJson_result()
    try:
      result.success = self._handler.executeJson(handler_ctx, args.sessionId, args.stmt)
    except:
      ex = sys.exc_info()[1]
      self._event_handler.handlerError(handler_ctx, 'executeJson', ex)
      result = Thrift.TApplicationException(message=repr(ex))
    return result

  @thrift_process_method(verifyClientVersion_args, oneway=False)
  def process_verifyClientVersion(self, args, handler_ctx):
    result = verifyClientVersion_result()
    try:
      result.success = self._handler.verifyClientVersion(handler_ctx, args.req)
    except:
      ex = sys.exc_info()[1]
      self._event_handler.handlerError(handler_ctx, 'verifyClientVersion', ex)
      result = Thrift.TApplicationException(message=repr(ex))
    return result

ContextIface._processor_type = ContextProcessor

fix_spec(all_structs)
del all_structs




#!/usr/bin/env python
# --coding:utf-8--

# Copyright (c) 2020 vesoft inc. All rights reserved.
#
# This source code is licensed under Apache 2.0 License,
# attached with Common Clause Condition 1.0, found in the LICENSES directory.

import logging


from nebula2.common.ttypes import ErrorCode
from nebula2.Exception import (
    AuthFailedException,
    IOErrorException,
    NotValidConnectionException,
    InValidHostname,
)

from nebula2.data.ResultSet import ResultSet

from nebula2.gclient.net.AuthResult import AuthResult
from nebula2.gclient.net.Session import Session
from nebula2.gclient.net.Connection import Connection
from nebula2.gclient.net.ConnectionPool import ConnectionPool

logging.basicConfig(level=logging.INFO, format='[%(asctime)s] %(levelname)-8s [%(filename)s:%(lineno)d]:%(message)s')



#!/usr/bin/env python
# --coding:utf-8--

# Copyright (c) 2020 vesoft inc. All rights reserved.
#
# This source code is licensed under Apache 2.0 License,
# attached with Common Clause Condition 1.0, found in the LICENSES directory.

import logging
import socket

from _thread import RLock
from nebula2.Exception import (
    InValidHostname,
    PartNotFoundException,
    SpaceNotFoundException,
    TagNotFoundException,
    EdgeNotFoundException
)
from nebula2.common.ttypes import HostAddr, ErrorCode
from nebula2.meta.ttypes import (
    HostStatus,
    ListTagsReq,
    ListEdgesReq,
    ListSpacesReq,
    GetPartsAllocReq,
    ListHostsReq,
    HostRole
)
from nebula2.meta import (
    ttypes,
    MetaService
)

from nebula2.fbthrift.transport import TSocket, TTransport
from nebula2.fbthrift.protocol import TBinaryProtocol


class MetaClient(object):
    def __init__(self, addresses, timeout):
        if len(addresses) == 0:
            raise RuntimeError('Input empty addresses')
        self._timeout = timeout
        self._connection = None
        self._retry_count = 3
        self._addresses = addresses
        for address in addresses:
            try:
                socket.gethostbyname(address[0])
            except Exception:
                raise InValidHostname(str(address[0]))
        self._leader = self._addresses[0]
        self._lock = RLock()

    def open(self):
        """open the connection to connect meta service

        :eturn: void
        """
        try:
            self.close()
            s = TSocket.TSocket(self._leader[0], self._leader[1])
            if self._timeout > 0:
                s.setTimeout(self._timeout)
            transport = TTransport.TBufferedTransport(s)
            protocol = TBinaryProtocol.TBinaryProtocol(transport)
            transport.open()
            self._connection = MetaService.Client(protocol)
        except Exception:
            raise

    def list_tags(self, space_id):
        """get all version tags

        :param space_id: the specified space id
        :eturn: list<TagItem>
        """
        with self._lock:
            if self._connection is None:
                raise RuntimeError('The connection is no open')
            req = ListTagsReq()
            req.space_id = space_id
            count = 0
            while count < self._retry_count:
                resp = self._connection.listTags(req)
                if resp.code != ErrorCode.SUCCEEDED:
                    if resp.code == ErrorCode.E_LEADER_CHANGED:
                        self.update_leader(resp.leader)
                        count = count + 1
                        continue
                    raise RuntimeError("List tags from space id:{} failed, error code: {}"
                                       .format(space_id, ErrorCode._VALUES_TO_NAMES.get(resp.code)))
                return resp.tags
            raise RuntimeError("List tags from space id:{} failed, error code: {}"
                               .format(space_id, ErrorCode._VALUES_TO_NAMES.get(resp.code)))

    def list_edges(self, space_id):
        """get all version edge

       :param space_id: the specified space id
       :eturn: list<EdgeItem>
       """
        with self._lock:
            if self._connection is None:
                raise RuntimeError('The connection is no open')
            req = ListEdgesReq()
            req.space_id = space_id
            count = 0
            while count < self._retry_count:
                resp = self._connection.listEdges(req)
                if resp.code != ErrorCode.SUCCEEDED:
                    if resp.code == ErrorCode.E_LEADER_CHANGED:
                        self.update_leader(resp.leader)
                        count = count + 1
                        continue
                    raise RuntimeError("List edges from space id:{} failed, error code: {}"
                                       .format(space_id, ErrorCode._VALUES_TO_NAMES.get(resp.code)))
                return resp.edges
            raise RuntimeError("List edges from space id:{} failed, error code: {}"
                               .format(space_id, ErrorCode._VALUES_TO_NAMES.get(resp.code)))

    def list_spaces(self):
        """get all spaces info

        :eturn: list<IdName>
        """
        with self._lock:
            if self._connection is None:
                raise RuntimeError('The connection is no open')
            req = ListSpacesReq()
            count = 0
            while count < self._retry_count:
                resp = self._connection.listSpaces(req)
                if resp.code != ErrorCode.SUCCEEDED:
                    if resp.code == ErrorCode.E_LEADER_CHANGED:
                        self.update_leader(resp.leader)
                        count = count + 1
                        continue
                    raise RuntimeError("List spaces failed, error code: {}"
                                       .format(ErrorCode._VALUES_TO_NAMES.get(resp.code)))
                return resp.spaces
            raise RuntimeError("List spaces failed, error code: {}"
                               .format(ErrorCode._VALUES_TO_NAMES.get(resp.code)))

    def list_hosts(self):
        """get all online hosts info

        :eturn: list<HostItem>
        """
        with self._lock:
            if self._connection is None:
                raise RuntimeError('The connection is no open')
            req = ListHostsReq()
            req.role = HostRole.STORAGE
            count = 0
            while count < self._retry_count:
                resp = self._connection.listHosts(req)
                if resp.code != ErrorCode.SUCCEEDED:
                    if resp.code == ErrorCode.E_LEADER_CHANGED:
                        self.update_leader(resp.leader)
                        count = count + 1
                        continue
                    raise RuntimeError("List spaces failed, error code: {}"
                                       .format(ErrorCode._VALUES_TO_NAMES.get(resp.code)))
                valid_hosts = []
                for host in resp.hosts:
                    if host.status == HostStatus.ONLINE:
                        valid_hosts.append(host)
                return valid_hosts
            raise RuntimeError("List spaces failed, error code: {}"
                               .format(ErrorCode._VALUES_TO_NAMES.get(resp.code)))

    def get_parts_alloc(self, space_id):
        """get all parts info of the specified space id

        :param space_id:
        :eturn: map<PartitionID, list<HostAddr>>
        """
        with self._lock:
            if self._connection is None:
                raise RuntimeError('The connection is no open')
            req = GetPartsAllocReq()
            req.space_id = space_id
            count = 0
            while count < self._retry_count:
                resp = self._connection.getPartsAlloc(req)
                if resp.code != ErrorCode.SUCCEEDED:
                    if resp.code == ErrorCode.E_LEADER_CHANGED:
                        self.update_leader(resp.leader)
                        count = count + 1
                        continue
                    raise RuntimeError("List parts from space id:{} failed, error code: {}"
                                       .format(space_id, ErrorCode._VALUES_TO_NAMES.get(resp.code)))
                return resp.parts
            raise RuntimeError("List parts from space id:{} failed, error code: {}"
                               .format(space_id, ErrorCode._VALUES_TO_NAMES.get(resp.code)))

    def close(self):
        """close the connection

        :eturn: void
        """
        try:
            if self._connection is not None:
                self._connection._iprot.trans.close()
        except Exception:
            raise

    def update_leader(self, leader):
        """update the leader meta info when hanppen leader change

        :param leader: the address of meta leader
        :eturn:
        """
        try:
            self._leader = (leader.host, leader.port)
            self.open()
        except Exception as e:
            logging.error(e)

    def __del__(self):
        self.close()


class MetaCache(object):
    class SpaceCache:
        def __init__(self):
            self.space_id = 0
            self.space_name = ''
            self.tag_items = {}
            self.edge_items = {}
            self.parts_alloc = {}

        def __repr__(self):
            return 'space_id: {}, space_name: {}, tag_items: {}, edge_items: {}, parts_alloc: {}'\
                .format(self.space_id,
                        self.space_name,
                        self.tag_items,
                        self.edge_items,
                        self.parts_alloc)

    def __init__(self, meta_addrs, timeout=2000, load_period=10, decode_type='utf-8'):
        self._decode_type = decode_type
        self._load_period = load_period
        self._lock = RLock()
        self._space_caches = {}
        self._space_id_names = {}
        self._storage_addrs = []
        self._storage_leader = {}
        self._close = False
        self._meta_client = MetaClient(meta_addrs, timeout)
        self._meta_client.open()

        # load meta data
        self._load_all()

    def close(self):
        """close the metaClient

        :eturn: void
        """
        if self._close:
            return
        self._close = True
        if self._meta_client is not None:
            self._meta_client.close()

    def __del__(self):
        self.close()

    def _load_all(self):
        """load all space info and schema info from meta services

        :eturn: void
        """
        try:
            spaces = self._meta_client.list_spaces()
            space_caches = {}
            space_id_names = {}
            for space in spaces:
                space_id = space.id.get_space_id()
                space_cache = MetaCache.SpaceCache()
                space_cache.space_id = space_id
                space_cache.space_name = space.name.decode('utf-8')
                space_id_names[space_id] = space_cache.space_name
                tags = self._meta_client.list_tags(space_id)
                edges = self._meta_client.list_edges(space_id)
                parts_alloc = self._meta_client.get_parts_alloc(space_id)
                for tag in tags:
                    tag_name = tag.tag_name.decode(self._decode_type)
                    if tag_name not in space_cache.tag_items.keys():
                        space_cache.tag_items[tag_name] = tag
                    else:
                        if space_cache.tag_items[tag_name].version < tag.version:
                            space_cache.tag_items[tag_name] = tag
                for edge in edges:
                    edge_name = edge.edge_name.decode(self._decode_type)
                    if edge_name not in space_cache.edge_items.keys():
                        space_cache.edge_items[edge_name] = edge
                    else:
                        if space_cache.edge_items[edge_name].version < edge.version:
                            space_cache.edge_items[edge_name] = edge
                    space_cache.edge_items[edge.edge_name.decode(self._decode_type)] = edge
                space_cache.parts_alloc = parts_alloc
                space_caches[space.name.decode(self._decode_type)] = space_cache

            hosts = self._meta_client.list_hosts()
            storage_addrs = []
            for host_item in hosts:
                storage_addrs.append(host_item.hostAddr)

            with self._lock:
                self._storage_addrs = storage_addrs
                self._space_caches = space_caches
                self._space_id_names = space_id_names
                for space_name in self._space_caches.keys():
                    if space_name in self._storage_leader.keys():
                        continue
                    parts_alloc = self._space_caches[space_name].parts_alloc
                    self._storage_leader[space_name] = {}
                    for part_id in parts_alloc:
                        self._storage_leader[space_name][part_id] = parts_alloc[part_id][0]
        except Exception as x:
            logging.error('Update meta data failed: {}'.format(x))
            import traceback
            logging.error(traceback.format_exc())

    def get_all_storage_addrs(self):
        """get all storage address

        :return: list[HostAddr]
        """
        return self._storage_addrs

    def get_tag_id(self, space_name, tag_name):
        """get tag id

        :param space_name:
        :param tag_name:
        :return: tag_id
        """
        with self._lock:
            tag_item = self._get_tag_item(space_name, tag_name)
            return tag_item.tag_id

    def get_edge_type(self, space_name, edge_name):
        """get edge type

        :param space_name:
        :param edge_name:
        :return: edge_type
        """
        with self._lock:
            edge_item = self._get_edge_item(space_name, edge_name)
            return edge_item.edge_type

    def get_space_id(self, space_name):
        """get space id

        :param space_name:
        :return: space_id
        """
        with self._lock:
            if space_name not in self._space_caches.keys():
                self._load_all()
                if space_name not in self._space_caches.keys():
                    raise SpaceNotFoundException(space_name)
            return self._space_caches[space_name].space_id

    def get_tag_schema(self, space_name, tag_name):
        """get tag schema

        :param space_name:
        :param tag_name:
        :return: schema
        """
        tag_item = self._get_tag_item(space_name, tag_name)
        return tag_item.schema

    def get_edge_schema(self, space_name, edge_name):
        """get edge schema

        :param space_name:
        :param edge_name:
        :return: schema
        """
        edge_item = self._get_edge_item(space_name, edge_name)
        return edge_item.schema

    def get_part_leader(self, space_name, part_id):
        """

        :param space_name:
        :param part_id:
        :return: storage ip port: HostAddr
        """
        part_leaders = self.get_part_leaders(space_name)
        if part_id not in part_leaders.keys():
            raise PartNotFoundException(part_id)
        return part_leaders[part_id]

    def get_part_leaders(self, space_name):
        """get all part leader info of the space

        :param space_name: space name
        :eturn: map<PartitionID, HostAddr>
        """
        with self._lock:
            if space_name not in self._storage_leader.keys():
                self._load_all()
                if space_name not in self._storage_leader.keys():
                    raise SpaceNotFoundException(space_name)
            return self._storage_leader[space_name]

    def get_part_alloc(self, space_name):
        """get all part info of the space

        :param space_name: space name
        :eturn: map<PartitionID, list<HostAddr>>
        """
        with self._lock:
            if space_name not in self._space_caches.keys():
                self._load_all()
                if space_name not in self._space_caches.keys():
                    raise SpaceNotFoundException(space_name)
            return self._space_caches[space_name].parts_alloc

    def _get_tag_item(self, space_name, tag_name):
        with self._lock:
            if space_name not in self._space_caches.keys():
                self._load_all()
                if space_name not in self._space_caches.keys():
                    raise SpaceNotFoundException(space_name)
            space_info = self._space_caches[space_name]
            if tag_name not in space_info.tag_items.keys():
                self._load_all()
                if tag_name not in space_info.tag_items.keys():
                    raise TagNotFoundException(tag_name)
            return space_info.tag_items[tag_name]

    def _get_edge_item(self, space_name, edge_name):
        with self._lock:
            if space_name not in self._space_caches.keys():
                self._load_all()
                if space_name not in self._space_caches.keys():
                    raise SpaceNotFoundException(space_name)
            space_info = self._space_caches[space_name]
            if edge_name not in space_info.edge_items.keys():
                self._load_all()
                if edge_name not in space_info.edge_items.keys():
                    raise EdgeNotFoundException(edge_name)
            return space_info.edge_items[edge_name]

    def update_storage_leader(self, space_id, part_id, address: HostAddr):
        """if the storage leader change, storage client need to call this function

        :param space_id:
        :param part_id:
        :param address: HostAddr, if the address is None, it means the leader can't connect,
        choose the peer as leader
        :return: coid
        """
        with self._lock:
            if space_id not in self._space_id_names.keys():
                logging.error("Space name:{} is not found".format(space_id))
                return
            space_name = self._space_id_names.get(space_id)
            if part_id not in self._storage_leader[space_name].keys():
                logging.error("part_id:{} is not found".format(space_name))
                return
            if address is not None:
                self._storage_leader[space_name][part_id] = address
                return
            part_addresses = self._space_caches[space_name].parts_alloc.get(part_id)
            for part_addr in part_addresses:
                if part_addr == address:
                    continue
                self._storage_leader[space_name][part_id] = part_addr





# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import, division, print_function, unicode_literals

from collections import OrderedDict, namedtuple

from nebula2.fbthrift.Thrift import TType


__all__ = ["create_client", "Serializer", "struct_to_dict", "parse_struct_spec"]
StructField = namedtuple("StructField", "id type name type_args default req_type")


def create_client(
    client_klass,
    host=None,
    port=None,
    client_type=None,
    path=None,
    timeout=None,
):
    """
    Given a thrift client class, and a host/port
    return a client using HeaderTransport
    """
    from nebula2.fbthrift.transport.TSocket import TSocket
    from nebula2.fbthrift.protocol.THeaderProtocol import THeaderProtocol

    sock = TSocket(host=host, port=port, unix_socket=path)
    sock.setTimeout(timeout)
    protocol = THeaderProtocol(
        sock,
        client_types=[client_type]
        if client_type
        else None,  # We accept the same as our inital send_
        client_type=client_type,  # Used for the inital send_
    )
    sock.open()
    return client_klass(protocol)


def parse_struct_spec(struct):
    """
    Given a thrift struct return a generator of parsed field information

    StructField fields:
        id - the field number
        type - a Thrift.TType
        name - the field name
        type_args - type arguments (ex: Key type Value type for maps)
        default - the default value
        req_type - the field required setting
            (0: Required, 1: Optional, 2: Optional IN, Required OUT)

    :param struct: a thrift struct
    :return: a generator of StructField tuples
    """
    for field in struct.thrift_spec:
        if not field:
            continue
        yield StructField._make(field)


def struct_to_dict(struct, defaults=False):
    """
    Given a Thrift Struct convert it into a dict
    :param struct: a thrift struct
    :param defaults: return default values
    :return: OrderedDict
    """
    adict = OrderedDict()
    union = struct.isUnion()
    if union and struct.field == 0:
        # if struct.field is 0 then it is unset escape
        return adict
    for field in parse_struct_spec(struct):
        if union:
            if field.id == struct.field:
                value = struct.value
            else:
                continue
        else:
            value = getattr(struct, field.name, field.default)
        if value != field.default or defaults:
            if field.type == TType.STRUCT:
                if value is not None:
                    sub_dict = struct_to_dict(value, defaults=defaults)
                    if sub_dict or defaults:  # Do not include empty sub structs
                        adict[field.name] = sub_dict
            elif field.type == TType.LIST:
                sub_list = __list_to_dict(value, field.type_args, defaults=defaults)
                if sub_list or defaults:
                    adict[field.name] = sub_list
            elif field.type == TType.SET:
                sub_set = __set_to_dict(value, field.type_args, defaults=defaults)
                if sub_set or defaults:
                    adict[field.name] = sub_set
            elif field.type == TType.MAP:
                sub_map = __map_to_dict(value, field.type_args, defaults=defaults)
                if sub_map or defaults:
                    adict[field.name] = sub_map
            else:
                adict[field.name] = value
        if union:  # If we got this far then we have the union value
            break
    return adict


def __list_to_dict(alist, type_args, defaults=False):
    """
    Given a python list-like collection, potentially containing Thrift Structs,
    convert it into a dict
    :param alist: a list or set
    :param defaults: return default values
    :return: List
    """
    if not alist:
        return alist

    element_type = type_args[0]
    if element_type == TType.STRUCT:
        return [struct_to_dict(element, defaults=defaults) for element in alist]
    if element_type == TType.LIST:
        return [
            __list_to_dict(element, type_args[1], defaults=defaults)
            for element in alist
        ]
    if element_type == TType.SET:
        return [
            __set_to_dict(element, type_args[1], defaults=defaults) for element in alist
        ]
    else:
        return alist


def __set_to_dict(aset, type_args, defaults=False):
    """
    Given a python set-like collection, potentially containing Thrift Structs
    and recursively parsing the elements
    :param aset: a set
    :param defaults: return default values
    :return: Set
    """
    if not aset:
        return aset

    element_type = type_args[0]
    if element_type == TType.STRUCT:
        return {struct_to_dict(element, defaults=defaults) for element in aset}
    if element_type == TType.LIST:
        return {
            __list_to_dict(element, type_args[1], defaults=defaults)
            for element in aset
        }
    if element_type == TType.SET:
        return {
            __set_to_dict(element, type_args[1], defaults=defaults) for element in aset
        }
    else:
        return aset


def __map_to_dict(amap, type_args, defaults=False):
    """
    Given a python dictionary, potentially containing Thrift Structs, convert it
    into a dict
    :param amap: a map
    :param defaults: return default values
    :return: Dict
    """
    if not amap:
        return amap

    keys, values = zip(*amap.items())

    keys = __list_to_dict(keys, type_args[:2], defaults=defaults)
    values = __list_to_dict(values, type_args[2:4], defaults=defaults)

    return dict(zip(keys, values))



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyre-unsafe

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from typing import Any, AnyStr, TypeVar  # noqa: F401

from nebula2.fbthrift.transport import TTransport
from nebula2.fbthrift.protocol import THeaderProtocol


# pyre-fixme[34]: `Variable[AnyStr <: [str, bytes]]` isn't present in the function's
#  parameters.
def serialize(protocol_factory, thr):
    # type: (Any, Any) -> AnyStr
    """Convenience method for serializing objects using the given
    protocol factory and a TMemoryBuffer."""
    transport = TTransport.TMemoryBuffer()
    protocol = protocol_factory.getProtocol(transport)
    thr.write(protocol)
    if isinstance(protocol, THeaderProtocol.THeaderProtocol):
        protocol.trans.flush()
    return transport.getvalue()


T = TypeVar("T")  # noqa: F401


def deserialize(protocol_factory, data, thr_out):
    # type: (Any, AnyStr, T) -> T
    """Convenience method for deserializing objects using the given
    protocol factory and a TMemoryBuffer.  returns its thr_out
    argument."""
    transport = TTransport.TMemoryBuffer(data)
    try:
        # pyre-fixme[16]: `T` has no attribute `thrift_spec`.
        protocol = protocol_factory.getProtocol(transport, thr_out.thrift_spec)  # noqa: T484
    except TypeError:
        protocol = protocol_factory.getProtocol(transport)
    if isinstance(protocol, THeaderProtocol.THeaderProtocol):
        # this reads the THeader headers to detect what the underlying
        # protocol is, as well as looking at transforms, etc.
        protocol.trans.readFrame(0)
        protocol.reset_protocol()
    # pyre-fixme[16]: `T` has no attribute `read`.
    thr_out.read(protocol)  # noqa: T484
    return thr_out



# --coding:utf-8--
#
# Copyright (c) 2021 vesoft inc. All rights reserved.
#
# This source code is licensed under Apache 2.0 License,
# attached with Common Clause Condition 1.0, found in the LICENSES directory.

import logging
import time

from nebula2.Exception import (
    IOErrorException,
    NotValidConnectionException,
)

from nebula2.data.ResultSet import ResultSet
from nebula2.gclient.net.AuthResult import AuthResult


class Session(object):
    def __init__(self, connection, auth_result: AuthResult, pool, retry_connect=True):
        self._session_id = auth_result.get_session_id()
        self._timezone_offset = auth_result.get_timezone_offset()
        self._connection = connection
        self._timezone = 0
        self._pool = pool
        self._retry_connect = retry_connect

    def execute(self, stmt):
        """execute statement

        :param stmt: the ngql
        :return: ResultSet
        """
        if self._connection is None:
            raise RuntimeError('The session has released')
        try:
            start_time = time.time()
            resp = self._connection.execute(self._session_id, stmt)
            end_time = time.time()
            return ResultSet(resp,
                             all_latency=int((end_time - start_time) * 1000000),
                             timezone_offset=self._timezone_offset)
        except IOErrorException as ie:
            if ie.type == IOErrorException.E_CONNECT_BROKEN:
                self._pool.update_servers_status()
                if self._retry_connect:
                    if not self._reconnect():
                        logging.warning('Retry connect failed')
                        raise IOErrorException(IOErrorException.E_ALL_BROKEN, ie.message)
                    resp = self._connection.execute(self._session_id, stmt)
                    end_time = time.time()
                    return ResultSet(resp,
                                     all_latency=int((end_time - start_time) * 1000000),
                                     timezone_offset=self._timezone_offset)
            raise
        except Exception:
            raise

    def execute_json(self, stmt):
        """execute statement and return the result as a JSON string
            Date and Datetime will be returned in UTC
            JSON struct:
            {
                "results": [
                {
                    "columns": [],
                    "data": [
                    {
                        "row": [
                        "row-data"
                        ],
                        "meta": [
                        "metadata"
                        ]
                    }
                    ],
                    "latencyInUs": 0,
                    "spaceName": "",
                    "planDesc ": {
                    "planNodeDescs": [
                        {
                        "name": "",
                        "id": 0,
                        "outputVar": "",
                        "description": {
                            "key": ""
                        },
                        "profiles": [
                            {
                            "rows": 1,
                            "execDurationInUs": 0,
                            "totalDurationInUs": 0,
                            "otherStats": {}
                            }
                        ],
                        "branchInfo": {
                            "isDoBranch": false,
                            "conditionNodeId": -1
                        },
                        "dependencies": []
                        }
                    ],
                    "nodeIndexMap": {},
                    "format": "",
                    "optimize_time_in_us": 0
                    },
                    "comment ": ""
                }
                ],
                "errors": [
                {
                    "code": 0,
                    "message": ""
                }
                ]
            }
        :param stmt: the ngql
        :return: JSON string
        """
        if self._connection is None:
            raise RuntimeError('The session has released')
        try:
            resp_json = self._connection.execute_json(self._session_id, stmt)
            return resp_json
        except IOErrorException as ie:
            if ie.type == IOErrorException.E_CONNECT_BROKEN:
                self._pool.update_servers_status()
                if self._retry_connect:
                    if not self._reconnect():
                        logging.warning('Retry connect failed')
                        raise IOErrorException(
                            IOErrorException.E_ALL_BROKEN, ie.message)
                    resp_json = self._connection.execute_json(
                        self._session_id, stmt)
                    return resp_json
            raise
        except Exception:
            raise

    def release(self):
        """release the connection to pool, and the session couldn't been use again

        :return:
        """
        if self._connection is None:
            return
        self._connection.signout(self._session_id)
        self._connection.is_used = False
        self._connection = None

    def ping(self):
        """check the connection is ok

        :return: True or False
        """
        if self._connection is None:
            return False
        return self._connection.ping()

    def _reconnect(self):
        try:
            self._connection.is_used = False
            conn = self._pool.get_connection()
            if conn is None:
                return False
            self._connection = conn
        except NotValidConnectionException:
            return False
        return True

    def __del__(self):
        self.release()
